var ControllerEditor = /** @class */ (function () {
    function ControllerEditor(controller, componentEditor) {
        this.controller = controller;
        this.componentEditor = componentEditor;
        this._currentRowModified = false;
    }
    Object.defineProperty(ControllerEditor.prototype, "currentRowModified", {
        get: function () {
            return this._currentRowModified;
        },
        set: function (modified) {
            if (modified !== this._currentRowModified) {
                this.controller.rowModifiedCallbacks.forEach(function (func) {
                    func(modified);
                });
                this._currentRowModified = modified;
            }
        },
        enumerable: true,
        configurable: true
    });
    ControllerEditor.prototype.validateEditors = function () {
        return this.componentEditor.validateEditors();
    };
    ControllerEditor.prototype.editDataRow = function (changedProp, newValue) {
        var _this = this;
        var oldValue = this.newRow[changedProp.objPropId];
        var propEditedPromise;
        if (this.setValue(changedProp, newValue)) {
            propEditedPromise = this.controller.afterPropEdited(changedProp);
        }
        else {
            propEditedPromise = this.controller.updateStatusAction;
        }
        // Manually wrap the action in a loader promise.
        // The afterPropEdited promise can't be set to the controllers currentAction because
        // the save or cancel action will be canceled when the user instantly presses the button
        // and the editDataRow wasn't executed for the edited column.
        if (propEditedPromise) {
            var timer = Utils.showLoaderWithDelay(250);
            propEditedPromise.always(function () {
                // Always clear the timeout since the prop is edited and the loader doesn't have to be displayed anymore.
                clearTimeout(timer);
                // Only hide the loader when there is no currentAction.
                // Otherwise the loader of a different action might get hidden.
                if (_this.controller.interfaceAvailable) {
                    Utils.hideLoader();
                }
            });
        }
        return propEditedPromise;
    };
    ControllerEditor.prototype.suggestCursorIndex = function (objProp) {
        this.componentEditor.suggestCursorIndex(objProp);
    };
    ControllerEditor.prototype.leave = function () {
        this.componentEditor.leave();
    };
    ControllerEditor.prototype.getActionTypeForProcessAction = function () {
        if (this.updateRow == null) {
            return 3 /* AddRow */;
        }
        else {
            return 4 /* EditRow */;
        }
    };
    ControllerEditor.prototype.saveRow = function () {
        var _this = this;
        return Utils.chainPromise(this.controller.updateStatusAction, function () {
            return _this.canLeaveEditorMode().then(function () {
                return _this.saveInternal();
            }).fail(function (objProp) {
                // When saveInternal fails, this code will also be called.
                if (objProp) {
                    _this.suggestCursorIndex(objProp);
                }
            });
        });
    };
    ControllerEditor.prototype.saveInternal = function () {
        var _this = this;
        var deferred = jQuery.Deferred();
        //Prevent saving when nothing was changed by the user
        if (!this.currentRowModified) {
            return deferred.resolve().promise();
        }
        var objView = this.controller.objView;
        var bookmark = TSFRowBookmark.fromObject(this.controller.obj, this.newRow);
        var successFul;
        var builder = RequestBuilderFactory.createRequestBuilder(this.controller.application, objView);
        builder.setCallbackMethod(function (reader, status) {
            if (status.success && (builder instanceof OData.RequestBuilder || reader && reader.length == 1)) {
                _this.currentRowModified = false;
                deferred.resolve(successFul(reader));
            }
            else {
                deferred.reject();
            }
        });
        if (this.updateRow == null) {
            successFul = function (reader) {
                // When refresh after mutation is true OR when it is an online table we should always update  the syncgroup
                if (objView.offlineRefreshAfterMutation || !objView.offlineAvailable) {
                    _this.controller.application.localSync.refreshSyncGroup(objView);
                }
                return TSFPkBookmark.fromReader(objView.obj, reader.item(0));
            };
            builder.create(bookmark);
        }
        else {
            successFul = function (reader) {
                // When refresh after mutation is true OR when it is an online table we should always update  the syncgroup
                if (objView.offlineRefreshAfterMutation || !objView.offlineAvailable) {
                    _this.controller.application.localSync.refreshSyncGroup(objView);
                }
                if (builder instanceof OData.RequestBuilder) {
                    // reader doesn't contain the pk on online subjects when updating so use the current data instead
                    return TSFPkBookmark.fromObject(objView.obj, _this.newRow);
                }
                else {
                    // Its important that the returned reader is used because it might contain resolved identities.
                    return TSFPkBookmark.fromReader(objView.obj, reader.item(0));
                }
            };
            builder.put(this.updateRow, bookmark);
        }
        return deferred.promise();
    };
    ControllerEditor.prototype.canLeaveEditorMode = function () {
        var sortedProps = this.controller.objView.getFormOrderedProps(), layoutState = this.controller.layoutState, fieldlayoutState = layoutState.fieldLayoutState, newRow = this.newRow;
        // The controller can't leave EditorMode when a prop is synchronizing
        // because a popup may be prompted that a value is invalid.
        // Also add an extra guard for the saveRowAccessType
        if (this.componentEditor.isSynchronizingProp ||
            layoutState.saveRowAccessType !== Enum.AccessType.Regular) {
            return jQuery.Deferred().reject(null).promise();
        }
        for (var i = 0; i < sortedProps.length; i++) {
            var objProp = sortedProps[i].objProp;
            var deferred = jQuery.Deferred();
            // The controller can't leave EditorMode when a prop is synchronizing
            // because a popup may be prompted that a value is invalid.
            // Also add an extra guard for the saveRowAccessType
            if (this.componentEditor.isSynchronizingProp ||
                layoutState.saveRowAccessType !== Enum.AccessType.Regular) {
                return jQuery.Deferred().reject(objProp);
            }
            if (fieldlayoutState[objProp.objPropId] == 4 /* Mandatory */ && (Utils.isEmpty(newRow[objProp.objPropId]))) {
                var transl = this.controller.application.language.transl(Model.TranslKeys.emptyMandatoryField);
                NavigationManager.showMessageWhenReady(this.controller.objView.objTransl.transl, [transl.replace("[objProp]", objProp.objPropTransl.translForm)]).always(function () {
                    deferred.reject(objProp);
                });
                return deferred.promise();
            }
        }
        return jQuery.Deferred().resolve().promise();
    };
    /**
     *
     * @param objProp The column for which to set the value for
     * @param newValue The new value
     * @return Returns wether the value changed or not
     */
    ControllerEditor.prototype.setValue = function (objProp, newValue) {
        var oldValue = this.newRow[objProp.objPropId];
        var valueChanged = false;
        if (objProp.xmlDttp == "date" && oldValue && newValue) {
            // Remove the milliseconds first
            if (!moment(this.newRow[objProp.objPropId]).milliseconds(0).isSame(newValue)) {
                valueChanged = true;
            }
        }
        else if (objProp.isNumeric() && !Utils.isEmpty(oldValue) && !Utils.isEmpty(newValue)) {
            // LocalSync might have saved doubles with an inconsistant precision.
            // Therefore parseFloat the existing value of the newRow to minify the precision.
            var parsed = parseFloat(oldValue);
            if (parsed != null && !isNaN(parsed) && parsed.toString() != newValue) {
                valueChanged = true;
            }
        }
        else if (objProp.isFileColumn && oldValue != null && newValue != null) {
            valueChanged = oldValue.FilePath != newValue.FilePath
                || oldValue.File != newValue.File;
        }
        else if (oldValue !== newValue) {
            valueChanged = true;
        }
        if (valueChanged) {
            this.newRow[objProp.objPropId] = newValue;
            this.currentRowModified = true;
        }
        return valueChanged;
    };
    ControllerEditor.prototype.toWebserviceObject = function (ignoreFileData) {
        if (ignoreFileData === void 0) { ignoreFileData = false; }
        var obj = this.controller.obj;
        return RowHelper.convertRowToWebserviceObject(obj, this.newRow, ignoreFileData, false);
    };
    ControllerEditor.prototype.initializeComponentEditor = function (component) {
        if (this.componentEditor && this.componentEditor.lazyInitialize) {
            this.componentEditor.lazyInitialize(component);
        }
    };
    ControllerEditor.prototype.createNewRow = function () {
        this.updateRow = null;
        var controller = this.controller;
        var objViewProps = this.controller.objView.objViewProps.array;
        this.newRow = {};
        // New row so this requires a save
        this.currentRowModified = true;
        for (var _i = 0, objViewProps_1 = objViewProps; _i < objViewProps_1.length; _i++) {
            var objViewProp = objViewProps_1[_i];
            if (objViewProp.defaultValue !== null) {
                this.newRow[objViewProp.objPropId] = objViewProp.defaultValue;
            }
        }
    };
    ControllerEditor.prototype.createEditRow = function () {
        var controller = this.controller;
        this.updateRow = controller.currentRowBookmark.toPkBookmark();
        var currentRow = new IndexedDb.IndexedDbDataReader(controller.currentRowBookmark.toObject());
        this.newRow = {};
        this.currentRowModified = false;
        for (var _i = 0, _a = controller.obj.objProps.array; _i < _a.length; _i++) {
            var column = _a[_i];
            RowHelper.copyValue(column, currentRow.getValueForObjProp(column), this.newRow);
        }
    };
    ControllerEditor.prototype.createCopyRow = function () {
        var controller = this.controller;
        this.updateRow = null;
        var currentRow = new IndexedDb.IndexedDbDataReader(controller.currentRowBookmark.toObject());
        this.newRow = {};
        // Copy is actually a new row and is therefore always modified
        this.currentRowModified = true;
        for (var _i = 0, _a = controller.obj.objProps.array; _i < _a.length; _i++) {
            var objProp = _a[_i];
            if (objProp.includeInCopy) {
                RowHelper.copyValue(objProp, currentRow.getValueForObjProp(objProp), this.newRow);
            }
        }
    };
    ControllerEditor.prototype.allowActions = function () {
        return this.componentEditor.allowActions();
    };
    ControllerEditor.prototype.isDefaultEditable = function () {
        return this.componentEditor.isDefaultEditable();
    };
    ControllerEditor.prototype.isCreatingNewRow = function () {
        return this.updateRow == null;
    };
    return ControllerEditor;
}());
