var Cube;
(function (Cube) {
    var ScreenComponent = /** @class */ (function () {
        function ScreenComponent(controller, contentElement) {
            this.controller = controller;
            this.contentElement = contentElement;
            // A list of deduced row dimensions.
            this.rowDimensions = [];
            // A list of deduced column dimensions.
            this.colDimensions = [];
            // A list of low level column dimensions. These are the columns that are at the lowest level of the aggregation.
            this.lowLevelColDimensions = [];
            this.highLevelRowDimensions = [];
            this.values = [];
            contentElement.classList.add("cube");
            // Use the default cube view.
            this.cubeModel = new Cube.CubeModel(controller.obj, controller.objView.objViewCube.defaultCubeView);
            this.cubeViewType = this.cubeModel.cubeView.defaultView;
            // Determine which view type should be used. Select chart if its selected otherwise use the CubeTable.
            if (this.cubeViewType == 2 /* Chart */) {
                this.component = new Cube.Graph(this, controller, contentElement);
            }
            else {
                this.component = new Cube.Table(this, controller, contentElement);
            }
        }
        /**
         * Bind controller events and append the data if the controller is initialized.
         */
        ScreenComponent.prototype.initialize = function () {
            var _this = this;
            if (this.controller.initialized) {
                var dataPromise = void 0;
                var defaultCubeViewId = this.cubeModel.cubeView.cubeViewId;
                if (this.controller.getCubeViewId() != defaultCubeViewId) {
                    dataPromise = this.controller.switchCubeView(defaultCubeViewId);
                }
                else {
                    dataPromise = this.controller.getData();
                }
                dataPromise.done(function (reader, status) {
                    _this.appendData(reader);
                }).fail(function (status) {
                    _this.component.invalidateData();
                });
            }
            this.controller.events.datasetChanged(function (args) {
                _this.appendData(args.dataset);
            });
        };
        /**
         * Clear the current cube and its information and rebuilds it according to the new data.
         * @param reader contains the data of the underlying datasource.
         */
        ScreenComponent.prototype.appendData = function (reader) {
            var _this = this;
            this.component.invalidateData();
            // Clear the saved dimensions.
            this.rowDimensions = [];
            this.colDimensions = [];
            // Determine the dimensions.
            for (var i = 0; i < reader.length; i++) {
                var item = reader.item(i);
                this.getDimensions(item);
            }
            // Calculate the maximum aggregation depth of the column dimensions.
            this.colDimensionAggregationDepth = Math.max.apply(Math, this.colDimensions.map(function (c) { return c.aggregationLevel; }));
            // Determine the column dimensions at the lowest level of the aggregation.
            this.lowLevelColDimensions = Array.prototype.filter.call(this.colDimensions, function (col) {
                return col.aggregationLevel == _this.colDimensionAggregationDepth;
            });
            this.rowDimensionAggregationDepth = Math.max.apply(Math, this.rowDimensions.map(function (c) { return c.aggregationLevel; }));
            this.highLevelRowDimensions = Array.prototype.filter.call(this.rowDimensions, function (row) {
                return row.aggregationLevel == 0;
            });
            this.component.appendData(reader);
        };
        /**
         * Determines and saves the dimensions that occur in a row.
         * @param item a row in the cube data.
         */
        ScreenComponent.prototype.getDimensions = function (item) {
            this.getGroupedDimensions(item, this.cubeModel.rows.array, this.rowDimensions);
            this.getGroupedDimensions(item, this.cubeModel.columns.array, this.colDimensions);
            var values = this.cubeModel.values.array;
            for (var i = 0; i < values.length; i++) {
                var valProp = values[i];
                if (this.values[valProp.objPropId] == null) {
                    var valVal = item.getValueForObjProp(valProp);
                    if (valVal != null) {
                        var val = new Cube.CubeValue();
                        val.id = valProp.objPropId;
                        val.value = valProp.objPropTransl.translGrid;
                        this.values[val.id] = val;
                        this.values.push(val);
                    }
                }
            }
        };
        ScreenComponent.prototype.getGroupedDimensions = function (item, props, dimensions) {
            // Dimensions are decuded from the objProps in the cube model.
            for (var i = 0; i < props.length; i++) {
                var dimProp = props[i];
                var dimVal = item.getValueForObjProp(dimProp);
                // Only create a dimension if there's actually a value.
                if (dimVal != null) {
                    var dimensionId = "tsf_col_";
                    var parentDimension = void 0;
                    // Add all parent dimension to the id.
                    for (var g = 0; g < i; g++) {
                        var parentDimProp = props[g];
                        var parentDimVal = item.getValueForObjProp(parentDimProp);
                        dimensionId += parentDimVal;
                        parentDimension = dimensions[dimensionId];
                    }
                    // Add the value to the dimension and add it the dimension to the dimension list.
                    dimensionId += dimVal;
                    if (dimensions[dimensionId] == null) {
                        var dim = new Cube.Dimension();
                        dim.id = dimensionId;
                        dim.aggregationLevel = i;
                        dim.value = dimVal;
                        dim.objProp = dimProp;
                        if (parentDimension != null) {
                            dim.parentDimensionId = parentDimension.id;
                            parentDimension.childDimensions.push(dim);
                        }
                        dimensions[dim.id] = dim;
                        dimensions.push(dim);
                    }
                }
            }
        };
        /**
         * Takes a row from the DataReader and determines the values for the row and column dimensions.
         * @param item a row in the cube data.
         * TODO: 1. Support for total rows.
         */
        ScreenComponent.prototype.addCubeValue = function (item) {
            var cubeRow = new Cube.CubeRow();
            var rows = this.cubeModel.rows.array;
            // Get all values for the row dimensions and add them to the cubeRow.
            for (var i = rows.length - 1; i >= 0; i--) {
                var rowProp = rows[i];
                var rowVal = item.getValueForObjProp(rowProp);
                if (rowVal != null) {
                    var rowDimensionId = "tsf_col_";
                    // Add all parent dimension to the id.
                    for (var g = 0; g < i; g++) {
                        var parentRowProp = rows[g];
                        var parentRowVal = item.getValueForObjProp(parentRowProp);
                        rowDimensionId += parentRowVal;
                    }
                    rowDimensionId += rowVal;
                    cubeRow.rowValues.add(rowDimensionId, rowVal);
                }
            }
            var columns = this.cubeModel.columns.array;
            for (var i = columns.length - 1; i >= 0; i--) {
                var colProp = columns[i];
                var colVal = item.getValueForObjProp(colProp);
                if (colVal != null) {
                    var columnDimensionId = "tsf_col_";
                    // Add all parent dimension to the id.
                    for (var g = 0; g < i; g++) {
                        var parentColProp = columns[g];
                        var parentColVal = item.getValueForObjProp(parentColProp);
                        columnDimensionId += parentColVal;
                    }
                    columnDimensionId += colVal;
                    cubeRow.colValues.add(columnDimensionId, colVal);
                }
            }
            var values = this.cubeModel.values.array;
            for (var j = 0; j < values.length; j++) {
                var dataProp = values[j];
                var dataVal = item.getValueForObjProp(dataProp);
                if (dataVal != null) {
                    cubeRow.valueId = dataProp.objPropId;
                    cubeRow.value = dataVal;
                    break;
                }
            }
            return cubeRow;
        };
        /**
        * method 'changeCubeView' Changes the used CubeView to build the CubeTable and CubeGraph.
        * @param cubeViewId states the id of the cubeViewId which should be used.
        */
        ScreenComponent.prototype.changeCubeView = function (cubeViewId) {
            var _this = this;
            // Only change the cubeView when it isn't the same as the currently used one.
            if (cubeViewId != null && this.cubeModel.cubeView.cubeViewId != cubeViewId) {
                // Clear the content
                this.invalidateData();
                var cubeView = this.controller.objView.objViewCube.cubeViews.key(cubeViewId);
                if (cubeView) {
                    // Create a new cubeModel with the new cubeView.
                    this.cubeModel = new Cube.CubeModel(this.controller.obj, cubeView);
                    // Re-append the data do not bind the dataset changed event because its still bound.
                    this.controller.switchCubeView(cubeView.cubeViewId).done(function (reader, status) {
                        _this.appendData(reader);
                    });
                }
            }
        };
        /**
        * method 'switchViewType' Switches between PivotTable and Chart view type.
        */
        ScreenComponent.prototype.switchViewType = function () {
            var _this = this;
            // Clear the content
            this.contentElement.innerHTML = "";
            // Set the viewType to the opposite CubeViewType and create the appropriate component.
            if (this.cubeViewType == 2 /* Chart */) {
                this.cubeViewType = 1 /* PivotTable */;
                this.component = new Cube.Table(this, this.controller, this.contentElement);
            }
            else {
                this.cubeViewType = 2 /* Chart */;
                this.component = new Cube.Graph(this, this.controller, this.contentElement);
            }
            // Re-append the data do not bind the dataset changed event because its still bound.
            if (this.controller.initialized) {
                this.controller.getData().done(function (reader, status) {
                    _this.appendData(reader);
                });
            }
        };
        /**
        * method 'invalidateData' Clears the current component and clear the saved dimensions.
        */
        ScreenComponent.prototype.invalidateData = function () {
            // Invalidate the current CubeComponent.
            this.component.invalidateData();
            // Clear the decuded dimensions.
            this.rowDimensions = [];
            this.colDimensions = [];
            this.values = [];
            this.lowLevelColDimensions = [];
        };
        /**
        * method 'dispose'
        */
        ScreenComponent.prototype.dispose = function () {
            // Nothing to dispose yet.
        };
        return ScreenComponent;
    }());
    Cube.ScreenComponent = ScreenComponent;
})(Cube || (Cube = {}));
