var IndexedDb;
(function (IndexedDb) {
    var RequestBuilder = /** @class */ (function () {
        function RequestBuilder(application, objView) {
            this.application = application;
            this.objView = objView;
            this.callbacks = [];
            this.wrapper = this.application.getStorageContainer();
        }
        RequestBuilder.fromTSFQuery = function (application, objView, query) {
            var builder = new IndexedDb.RequestBuilder(application, objView);
            builder.query = query;
            return builder;
        };
        RequestBuilder.prototype.setCallbackMethod = function (callback) {
            this.callbacks.push(callback);
            return this;
        };
        RequestBuilder.prototype.getEmptyReader = function () {
            return new IndexedDb.IndexedDbArrayReader([]);
        };
        RequestBuilder.prototype.abortRequest = function () {
            //Not implemented for now as it's a bigger deal to abort HTTP requests than it is to abort IndexedDB requests.
        };
        RequestBuilder.prototype.get = function () {
            var _this = this;
            if (this.query) {
                var obj = this.objView !== null ? this.objView.obj : null;
                this.wrapper.executeQuery(this.query, obj, this.objView).done(function (result) {
                    _this.fireCallbacks(result, new RequestStatus(true));
                });
            }
        };
        RequestBuilder.prototype.create = function (newRow) {
            this.createInternal(newRow, true);
        };
        RequestBuilder.prototype.createInternal = function (newRow, markForOnlineSync) {
            var _this = this;
            this.wrapper.resolveIdentitiesInRow(this.objView, newRow).always(function (updatedRow) {
                var promise = _this.wrapper.insertBookmark(_this.objView.obj, updatedRow, markForOnlineSync);
                promise.done(function (reader) {
                    var status = new RequestStatus((reader != null));
                    if (!status.success) {
                        status.messages.push("Error creating record");
                    }
                    var offlineLogic = Utils.getOfflineLogicObject(_this.application, _this.objView.obj);
                    if (offlineLogic && offlineLogic.executeAfterInsert) {
                        offlineLogic.executeAfterInsert(_this.createOfflineContext(), updatedRow.toObject(), function (ret) {
                            // Just use the existing reader, a deleted row can't be modified.
                            _this.fireCallbacks(new IndexedDb.IndexedDbArrayReader([ret]), new RequestStatus(ret != null));
                        });
                    }
                    else {
                        _this.fireCallbacks(reader, status);
                    }
                });
                promise.fail(function (reader) {
                    var status = new RequestStatus(false);
                    status.messages.push("Error creating record");
                    _this.fireCallbacks(null, status);
                });
            });
        };
        RequestBuilder.prototype.put = function (referenceContract, newRow) {
            this.putInternal(referenceContract, newRow, true);
        };
        RequestBuilder.prototype.putInternal = function (referenceContract, newRow, markForOnlineSync) {
            var _this = this;
            // This may seem inefficient(and it is), but its important that both the reference contract and the newRow get resolved.
            this.wrapper.resolveIdentitiesInRow(this.objView, referenceContract).always(function (updatedReferenceContract) {
                _this.wrapper.resolveIdentitiesInRow(_this.objView, newRow).always(function (updatedRow) {
                    var promise = _this.wrapper.updateRow(_this.objView.obj, updatedReferenceContract, updatedRow, markForOnlineSync);
                    promise.done(function (reader) {
                        var status = new RequestStatus(true);
                        var offlineLogic = Utils.getOfflineLogicObject(_this.application, _this.objView.obj);
                        if (offlineLogic && offlineLogic.executeAfterUpdate) {
                            offlineLogic.executeAfterUpdate(_this.createOfflineContext(), updatedReferenceContract.toObject(), updatedRow.toObject(), function (ret) {
                                // Just use the existing reader, a deleted row can't be modified.
                                _this.fireCallbacks(new IndexedDb.IndexedDbArrayReader([ret]), new RequestStatus(ret != null));
                            });
                        }
                        else {
                            _this.fireCallbacks(reader, status);
                        }
                    });
                    promise.fail(function (message) {
                        var status = new RequestStatus(false);
                        status.messages.push("Error updating record: " + message);
                        console.log(status.messages);
                        _this.fireCallbacks(null, status);
                    });
                });
            });
        };
        // Delete is uppercase, the android browser (on the emulator) gives a syntax error when lowercase because delete is a keyword.
        RequestBuilder.prototype.Delete = function (row) {
            this.deleteInternal(row, true);
        };
        RequestBuilder.prototype.deleteInternal = function (row, markForOnlineSync) {
            var _this = this;
            this.wrapper.resolveIdentitiesInRow(this.objView, row).always(function (updatedRow) {
                // TODO: Why does deleteRow return an DataArrayReader with pk bookmarks? Only one bookmark would suffice.
                var promise = _this.wrapper.deleteRow(_this.objView.obj, updatedRow.toPkBookmark(), markForOnlineSync);
                promise.done(function (reader) {
                    var status = new RequestStatus(true);
                    var offlineLogic = Utils.getOfflineLogicObject(_this.application, _this.objView.obj);
                    if (offlineLogic && offlineLogic.executeAfterDelete) {
                        offlineLogic.executeAfterDelete(_this.createOfflineContext(), updatedRow.toObject(), function (ret) {
                            _this.fireCallbacks(new IndexedDb.IndexedDbArrayReader([ret]), new RequestStatus(ret != null));
                        });
                    }
                    else {
                        _this.fireCallbacks(reader, status);
                    }
                });
                promise.fail(function (message) {
                    var status = new RequestStatus(false);
                    status.messages.push("Error deleting record: " + message);
                    console.log(status.messages);
                    _this.fireCallbacks(null, status);
                });
            });
        };
        RequestBuilder.prototype.fireCallbacks = function (reader, status) {
            for (var i = 0; i < this.callbacks.length; i++) {
                var callback = this.callbacks[i];
                callback(reader, status);
            }
            // Clear subscribed callbacks for this call.
            this.callbacks = [];
        };
        RequestBuilder.prototype.createOfflineContext = function () {
            return {
                application: this.application,
                obj: this.objView.obj,
                user: Model.TSFApplicationManager.Current.user.username,
                synchronize: this.application.localSync.fullSynchronize
            };
        };
        RequestBuilder.prototype.executeLayout = function (layoutContract, layoutmode, importMode) {
            var _this = this;
            // Ignore the importMode since its never true for offline logic.
            var offlineLogic = Utils.getOfflineLogicObject(this.application, this.objView.obj);
            if (offlineLogic && offlineLogic.executeLayout) {
                offlineLogic.executeLayout(this.createOfflineContext(), layoutmode, layoutContract, function (ret) {
                    _this.fireCallbacks(new IndexedDb.IndexedDbArrayReader([ret]), new RequestStatus(ret != null));
                });
            }
            else {
                this.fireCallbacks(new IndexedDb.IndexedDbArrayReader([layoutContract]), new RequestStatus(true));
            }
        };
        RequestBuilder.prototype.executeDefault = function (defaultContract, layoutmode, importMode, from) {
            var _this = this;
            // Ignore the importMode since its never true for offline logic.
            var offlineLogic = Utils.getOfflineLogicObject(this.application, this.objView.obj);
            if (offlineLogic && offlineLogic.executeDefault) {
                if (from) {
                    jQuery.extend(defaultContract, { "cursor_from_col_id": from.objPropId });
                }
                offlineLogic.executeDefault(this.createOfflineContext(), layoutmode, defaultContract, function (ret) {
                    _this.fireCallbacks(new IndexedDb.IndexedDbArrayReader([ret]), new RequestStatus(ret != null));
                });
            }
            else {
                this.fireCallbacks(new IndexedDb.IndexedDbArrayReader([defaultContract]), new RequestStatus(true));
            }
        };
        RequestBuilder.prototype.executeContext = function (contextContract) {
            var _this = this;
            var offlineLogic = Utils.getOfflineLogicObject(this.application, this.objView.obj);
            if (offlineLogic && offlineLogic.executeContext) {
                offlineLogic.executeContext(this.createOfflineContext(), contextContract, function (ret) {
                    _this.fireCallbacks(new IndexedDb.IndexedDbArrayReader([ret]), new RequestStatus(ret != null));
                });
            }
            else {
                this.fireCallbacks(new IndexedDb.IndexedDbArrayReader([contextContract]), new RequestStatus(true));
            }
        };
        RequestBuilder.prototype.executeBadge = function (objView, rowObject) {
            var _this = this;
            var offlineLogic = Utils.getOfflineLogicObject(this.application, objView.obj);
            if (objView.objViewId != null) {
                rowObject = jQuery.extend({ VariantId: objView.objViewId }, rowObject);
            }
            if (offlineLogic && offlineLogic.executeBadge) {
                offlineLogic.executeBadge(this.createOfflineContext(), rowObject, function (ret) {
                    _this.fireCallbacks(new IndexedDb.IndexedDbArrayReader([ret]), new RequestStatus(ret != null));
                });
            }
            else {
                this.fireCallbacks(new IndexedDb.IndexedDbArrayReader([{}]), new RequestStatus(true));
            }
        };
        RequestBuilder.prototype.offlineTaskAvailable = function () {
            var offlineLogic = Utils.getOfflineLogicObject(this.application, this.objView.obj);
            return offlineLogic && offlineLogic.executeTask !== undefined;
        };
        RequestBuilder.prototype.executeOfflineTask = function (contract) {
            var _this = this;
            var offlineLogic = Utils.getOfflineLogicObject(this.application, this.objView.obj);
            offlineLogic.executeTask(this.createOfflineContext(), contract, function (ret) {
                _this.fireCallbacks(new IndexedDb.IndexedDbArrayReader([ret]), new RequestStatus(ret != null));
            });
        };
        RequestBuilder.prototype.executeTask = function (contract) {
            var _this = this;
            this.wrapper.queueTask(TSFRowBookmark.fromObject(this.objView.obj, contract)).done(function () {
                _this.fireCallbacks(new IndexedDb.IndexedDbArrayReader([contract]), new RequestStatus(true));
            }).fail(function () {
                _this.fireCallbacks(new IndexedDb.IndexedDbArrayReader([contract]), new RequestStatus(false));
            });
        };
        RequestBuilder.prototype.allowCustomIdentitiesInContracts = function () {
            return true;
        };
        return RequestBuilder;
    }());
    IndexedDb.RequestBuilder = RequestBuilder;
})(IndexedDb || (IndexedDb = {}));
