var IndexedDb;
(function (IndexedDb) {
    function fixIOSShim() {
        // This happens on iOS 9, the shim couldn't overwrite it so it stays null.
        // Update IDBKeyRange when this happens, we can't fix the indexedDB property.
        // iOS 10 implementation is buggy - adding records to a store, then call clear, the items arent in the store but when adding them again and walking the index
        // they appear multiple times - so we force the shim there (which does work again in iOS 10 :))
        var iOSVersion = Utils.getiOSVersion();
        if (iOSVersion >= 10) {
            window["shimIndexedDB"].__useShim();
        }
        if (!indexedDB || (window.indexedDB === window["shimIndexedDB"])) {
            window["IDBKeyRange"] = window["idbModules"].IDBKeyRange;
        }
    }
    IndexedDb.fixIOSShim = fixIOSShim;
    function open(name, version) {
        // Setting the shim to window.indexedDB doesn't work on iOS.. so this is a crappy workaround.  (iOS 9)
        if (indexedDB && indexedDB.open) {
            return indexedDB.open(name, version);
        }
        else {
            return window["shimIndexedDB"].open(name, version);
        }
    }
    IndexedDb.open = open;
    function deleteDatabase(name) {
        // Setting the shim to window.indexedDB doesn't work on iOS.. so this is a crappy workaround. (iOS 9)
        if (indexedDB && indexedDB.deleteDatabase) {
            return indexedDB.deleteDatabase(name);
        }
        else {
            return window["shimIndexedDB"].deleteDatabase(name);
        }
    }
    IndexedDb.deleteDatabase = deleteDatabase;
    function deleteDatabasePromise(key) {
        var deferred = jQuery.Deferred();
        var req = IndexedDb.deleteDatabase(key);
        req.onsuccess = function (evt) {
            deferred.resolve();
        };
        req.onerror = function (evt) {
            deferred.reject();
        };
        return deferred.promise();
    }
    IndexedDb.deleteDatabasePromise = deleteDatabasePromise;
    function addIndexes(obj, row) {
        // Add all the tsf_ columns to the given row.
        addPKIndexes(obj, row);
        addSortIndex(obj, row);
        addLookupFilterIndexes(obj, row);
        addObjLinkIndexes(obj, row);
    }
    IndexedDb.addIndexes = addIndexes;
    function addPKIndexes(obj, row) {
        if (obj.pkProps.length != 1) {
            var value = "";
            for (var i = 0; i < obj.pkProps.length; i++) {
                if (i != 0)
                    value += "-";
                value += row[obj.pkProps[i].objPropId];
            }
            row[IndexedDb.IndexedDbWrapper.pkIndexName] = value;
        }
    }
    function addSortIndex(obj, row) {
        for (var sortIndexName in obj.sortIndexes) {
            var sortIndex = obj.sortIndexes[sortIndexName];
            var sortProps = sortIndex.sortingProps;
            if (sortProps.length > 1 || sortProps[0].objProp.isNumeric() || !sortProps[0].mandatory) {
                var value = "";
                for (var i = 0; i < sortProps.length; i++) {
                    var sortViewProp = sortProps[i], sortObjProp = sortViewProp.objProp;
                    if (sortObjProp.isNumeric()) {
                        value += ('00000000000000' + row[sortObjProp.objPropId]).slice(-15);
                    }
                    else if (sortObjProp.xmlDttp == "date") {
                        // We need to special case date's so the sorting is correct (as it is converted to strings)
                        if (row[sortObjProp.objPropId] == undefined) {
                            value += "|";
                        }
                        else {
                            value += moment(row[sortObjProp.objPropId]).toJSON();
                        }
                    }
                    else if (!sortViewProp.mandatory) {
                        if (row[sortObjProp.objPropId] == undefined) {
                            // Actually a bit random but this character comes after the alphanumeric characters
                            value += "|";
                        }
                        else {
                            value += row[sortObjProp.objPropId];
                        }
                    }
                    else {
                        value += row[sortObjProp.objPropId];
                    }
                }
                row[sortIndex.keyPath] = value;
            }
        }
    }
    function addLookupFilterIndexes(obj, row) {
        for (var lookupIndexName in obj.lookupFilterIndexes) {
            var lookupIndex = obj.lookupFilterIndexes[lookupIndexName];
            if (lookupIndex.lookupFilters.length > 1) {
                var value = "";
                for (var i = 0; i < lookupIndex.lookupFilters.length; i++) {
                    if (i != 0)
                        value += "-";
                    value += row[lookupIndex.lookupFilters[i].targetProp.objPropId];
                }
                row[lookupIndex.keyPath] = (lookupIndex.upperCase ? value.toUpperCase() : value);
            }
            else {
                if (!obj.lookupDisplay.lookUpDisplayColId)
                    continue;
                var value = row[obj.lookupDisplay.lookUpDisplayColId];
                if (value != null && lookupIndex.upperCase) {
                    row[lookupIndex.keyPath] = value.toUpperCase();
                }
            }
        }
    }
    function addObjLinkIndexes(obj, row) {
        for (var objLinkIndexName in obj.objLinkIndexes) {
            var objLinkIndex = obj.objLinkIndexes[objLinkIndexName];
            if (objLinkIndex.objLinkProps.length > 1) {
                var value = "";
                for (var i = 0; i < objLinkIndex.objLinkProps.length; i++) {
                    if (i != 0)
                        value += "-";
                    value += row[objLinkIndex.objLinkProps[i].targetProp.objPropId];
                }
                row[objLinkIndex.indexName] = value;
            }
        }
    }
    function createPKKeyRangeFromBookmark(obj, bookmark) {
        var bookmarkValues = bookmark.conditions
            .slice()
            .sort(function (a, b) { return obj.pkProps.indexOf(a.objProp) - obj.pkProps.indexOf(b.objProp); })
            .filter(function (con) { return obj.pkProps.indexOf(con.objProp) !== -1; })
            .map(function (con) { return con.value; });
        return createKeyRangeFromValues(obj, bookmarkValues);
    }
    IndexedDb.createPKKeyRangeFromBookmark = createPKKeyRangeFromBookmark;
    function createKeyRangeFromValues(obj, valueArray) {
        if (valueArray.length == 1) {
            return IDBKeyRange.only(valueArray[0]);
        }
        else {
            return IDBKeyRange.only(valueArray.join("-"));
        }
    }
    IndexedDb.createKeyRangeFromValues = createKeyRangeFromValues;
})(IndexedDb || (IndexedDb = {}));
IndexedDb.fixIOSShim();
