var IndexedDb;
(function (IndexedDb) {
    var IndexedDbWrapper = /** @class */ (function () {
        function IndexedDbWrapper(application) {
            this.application = application;
            this.isDisposed = false;
        }
        IndexedDbWrapper.prototype.getDatabase = function () {
            return this.db;
        };
        IndexedDbWrapper.prototype.initializeDatabase = function () {
            if (this.db)
                return jQuery.Deferred().resolve().promise();
            if (window["idbModules"] && window.indexedDB === window["idbModules"].shimIndexedDB) {
                console.log('IndexedDb shim is used');
            }
            return this.openDatabase();
        };
        IndexedDbWrapper.prototype.openDatabase = function () {
            var deferred = jQuery.Deferred();
            var that = this, application = this.application;
            var applString = application.projectName + " - " + application.guiApplId;
            var request;
            try {
                request = IndexedDb.open(application.guiApplId, 3);
            }
            catch (error) {
                var errorMsg = application.language.transl(Model.TranslKeys.errorOpeningDatabase).replace("[db_name]", applString);
                // You might expect only the QuotaExceededError but that is not the error safari is throwing
                if (error.name == "SecurityError" || error.name == "QuotaExceededError") {
                    deferred.reject(errorMsg + ("\r\n" + application.language.transl(Model.TranslKeys.insufficientSpaceAvailable)));
                }
                else {
                    deferred.reject(errorMsg + ("\r\n" + (error.message || error.name)));
                }
                return deferred.promise();
            }
            // Events are set in chronological order, first upgrade will be raised and then error or success
            request.onupgradeneeded = function (event) {
                var db = event.target.result, objs = application.model.objs.array;
                // createObjectStore calls are synchronous.
                for (var i = 0; i < objs.length; i++) {
                    if (event.oldVersion == 0) {
                        var obj = objs[i], store = db.createObjectStore(obj.objId, { keyPath: 'tsf_id', autoIncrement: true });
                        // Add pk, lookup filter, lookup translation and obj link indexes.
                        store.createIndex(obj.pkIndex.indexName, obj.pkIndex.keyPath);
                        for (var sortIndexName in obj.sortIndexes) {
                            var sortIndex = obj.sortIndexes[sortIndexName];
                            store.createIndex(sortIndex.indexName, sortIndex.keyPath);
                        }
                        for (var lookupIndexName in obj.lookupFilterIndexes) {
                            var lookupIndex = obj.lookupFilterIndexes[lookupIndexName];
                            store.createIndex(lookupIndex.indexName, lookupIndex.keyPath);
                        }
                        for (var objLinkIndexName in obj.objLinkIndexes) {
                            var objLinkIndex = obj.objLinkIndexes[objLinkIndexName];
                            store.createIndex(objLinkIndex.indexName, objLinkIndex.keyPath);
                        }
                    }
                }
                if (event.oldVersion == 0) {
                    db.createObjectStore(IndexedDbWrapper.queueTableName, { keyPath: 'tsf_id', autoIncrement: true });
                    db.createObjectStore(IndexedDbWrapper.identityTableName, { keyPath: 'id', autoIncrement: true });
                    db.createObjectStore(IndexedDbWrapper.fileTableName, { keyPath: "HashCode" });
                    db.createObjectStore(IndexedDbWrapper.fileRefTableName, { keyPath: "HashCode" });
                }
                if (event.oldVersion < 3) {
                    db.createObjectStore(IndexedDbWrapper.responseBufferTableName, { keyPath: 'obj_id' });
                    db.createObjectStore(IndexedDbWrapper.syncGroupLastSyncTableName, { keyPath: 'sync_group_id' });
                }
                console.log('upgrade end: ' + applString);
            };
            request.onerror = function (event) {
                var errorMessage = "Error creating database: " + applString + "\r\n" + request.error.message;
                deferred.reject(errorMessage);
            };
            request.onblocked = request.onerror;
            request.onsuccess = function (event) {
                that.db = request.result;
                deferred.resolve();
            };
            return deferred.promise();
        };
        IndexedDbWrapper.prototype.createFilterHelper = function (query) {
            var filter = new FilterHelper.WhereClauseHelper();
            IndexedDbWrapper.createFilterHelperInner(filter, query.whereClause.contextFilters);
            IndexedDbWrapper.createFilterHelperPrefilters(filter, query.whereClause.prefilterState);
            IndexedDbWrapper.createFilterHelperInner(filter, query.whereClause.userFilters);
            IndexedDbWrapper.createFilterHelperInner(filter, query.whereClause.combinedFilterConditions, 1 /* Or */);
            return filter;
        };
        IndexedDbWrapper.createFilterHelperPrefilters = function (filter, prefilterState) {
            if (!prefilterState) {
                return;
            }
            for (var i = 0; i < prefilterState.length; i++) {
                var prefilter = prefilterState[i];
                if (prefilter.selected && prefilter.prefilter.prefilterProps.length > 0) {
                    var prefilterProps = prefilter.prefilter.prefilterProps;
                    for (var j = 0; j < prefilterProps.length; j++) {
                        var prefilterProp = prefilterProps[j];
                        var condition = prefilterProp.filterCondition;
                        if (prefilterProp.viewProp.objProp.xmlDttp == "date") {
                            if (condition == Enum.FilterCondition.EqualTo) {
                                condition = Enum.FilterCondition.DateEqualTo;
                            }
                            else if (condition == Enum.FilterCondition.NotEqualTo) {
                                condition = Enum.FilterCondition.DateNotEqualTo;
                            }
                        }
                        var comparison = Enum.FilterCondition[condition];
                        comparison = comparison.replace(/(^)./g, function (a) {
                            return a.toLowerCase();
                        });
                        if (!FilterHelper.Where.supportsFilterCondition(comparison)) {
                            console.log("Unsupported condition: " + comparison);
                            continue;
                        }
                        filter.where(prefilterProp.viewProp.objPropId, false, 0 /* And */)[comparison](prefilterProp.filterFromValue, prefilterProp.filterUntilValue);
                    }
                }
            }
        };
        IndexedDbWrapper.createFilterHelperInner = function (filter, conditions, operator) {
            if (operator === void 0) { operator = 0 /* And */; }
            for (var i = 0; i < conditions.length; i++) {
                var condition = conditions[i];
                var comparison = IndexedDbWrapper.filterConditionToString(condition.comparision);
                filter.where(condition.targetPropId, condition.ignoreCasing, operator)[comparison](condition.dbFilterValue, condition.dbFilterUntilValue);
            }
        };
        IndexedDbWrapper.filterConditionToString = function (condition) {
            var comparison = Enum.FilterCondition[condition];
            comparison = comparison.replace(/(^)./g, function (a) {
                return a.toLowerCase();
            });
            return comparison;
        };
        IndexedDbWrapper.prototype.executeQuery = function (query, obj, objView) {
            var _this = this;
            if (objView === void 0) { objView = null; }
            var deferred = jQuery.Deferred();
            var objectStore = this.getObjectStore(obj.objId, "readonly");
            if (!objectStore) {
                return deferred.resolve().promise();
            }
            var resultPromise;
            // Execute the query as a batched query when it has one or more orred conditions.
            if (query.whereClause.orredConditions.length > 0) {
                resultPromise = this.executeBatchedQuery(objectStore, query, obj);
            }
            else {
                var filterHelper = this.createFilterHelper(query);
                resultPromise = filterHelper.filterStore(objectStore, query, objView, this);
            }
            resultPromise.then(function (result) {
                // Don't manually sort and page when there is a preferredIndex but no keyRange.
                // preferredIndexes without keyRanges are used when the pk equals the start of the sorting.
                if (!query.preferredIndex || query.keyRange) {
                    if (query.currentSorting && query.currentSorting.length > 0) {
                        result = _this.sortArray(result, query, obj);
                    }
                    if (query.usePagination) {
                        result = _this.paging(result, query);
                    }
                }
                else if (query.currentSorting && query.currentSorting.length > 0) {
                    var containsStringSorting = query.currentSorting.filter(function (sortingContract) {
                        //Get the objProp the sorting is taking place on.
                        var sortProp = Array.prototype.filter.call(obj.objProps.array, function (prop) { return prop.webserviceId == sortingContract.name; });
                        return sortProp.length > 0 && sortProp[0].xmlDttp === "string";
                    }).length > 0;
                    //Always use our sort when there's a string in the sorted columns,
                    //because IndexedDb sorts case sensitively which we don't want. (TCP 52083)
                    if (containsStringSorting) {
                        result = _this.sortArray(result, query, obj);
                    }
                }
                deferred.resolve(new IndexedDb.IndexedDbArrayReader(result));
            });
            return deferred.promise();
        };
        IndexedDbWrapper.prototype.executeBatchedQuery = function (objectStore, query, obj) {
            var deferred = jQuery.Deferred(), resolveList = [];
            var resultArr = [];
            // Execute a query for every or condition.
            // Collect all results and return them as one array.
            for (var i = 0; i < query.whereClause.orredConditions.length; i++) {
                var filterHelper = this.createFilterHelper(query);
                var orCondition = query.whereClause.orredConditions[i];
                // fix filter conditions, last item is the filter condition which we should use as filter
                // the other ones are for the keyrange (project-subproject) when we want the row for activity
                var keyRangeConditions = [];
                if (orCondition.length === 1) {
                    keyRangeConditions.push(orCondition[0].dbFilterValue);
                }
                else {
                    for (var j = 0; j < orCondition.length - 1; j++) {
                        keyRangeConditions.push(orCondition[j].dbFilterValue);
                    }
                }
                query.keyRange = IndexedDb.createKeyRangeFromValues(obj, keyRangeConditions);
                var lastCondition = orCondition[orCondition.length - 1];
                var comparison = IndexedDbWrapper.filterConditionToString(lastCondition.comparision);
                filterHelper.where(lastCondition.targetPropId, lastCondition.ignoreCasing, 0 /* And */)[comparison](lastCondition.dbFilterValue, lastCondition.dbFilterUntilValue);
                resolveList.push(filterHelper.filterStore(objectStore, query, null, null).then(function (result) {
                    // Concat the global return array with the result this particular query.
                    resultArr = resultArr.concat(result);
                }));
            }
            // Always resolve, the filterStore doesn't reject yet.
            jQuery.when.apply(undefined, resolveList).always(function () { deferred.resolve(resultArr); });
            return deferred.promise();
        };
        IndexedDbWrapper.prototype.sortArray = function (arr, query, obj) {
            // This is modified version of http://jsperf.com/multi-column-sort/5 (multiColumnSort4)
            var sorting = query.currentSorting;
            var functionBody = '';
            for (var i = 0; i < sorting.length; i++) {
                var sort = sorting[i], sortingName = sort.name, direction = sort.direction;
                var sortProp = Array.prototype.filter.call(obj.objProps.array, function (prop) { return prop.objPropId == sortingName; });
                var isNumeric = (sortProp.length > 0 && sortProp[0].isNumeric());
                var isString = (sortProp.length > 0 && sortProp[0].xmlDttp === "string");
                if (isNumeric) {
                    functionBody += "if (parseFloat(a." + sortingName + ") " + (direction == Enum.SortDirection.Ascending ? ">" : "<") + " parseFloat(b." + sortingName + ")) return 1;";
                }
                else if (isString) {
                    functionBody += "if (a. " + sortingName + " === undefined || a. " + sortingName + " === null) return " + (direction === Enum.SortDirection.Ascending ? "0" : "1") + ";";
                    functionBody += "if (b. " + sortingName + " === undefined || b. " + sortingName + " === null) return " + (direction === Enum.SortDirection.Ascending ? "1" : "0") + ";";
                    functionBody += "if (a." + sortingName + ".toUpperCase() " + (direction === Enum.SortDirection.Ascending ? ">" : "<") + " b." + sortingName + ".toUpperCase()) return 1;";
                }
                else {
                    functionBody += "if (a." + sortingName + " " + (direction === Enum.SortDirection.Ascending ? ">" : "<") + " b." + sortingName + ") return 1;";
                }
                functionBody += "if (a." + sortingName + " === b." + sortingName + ") ";
                functionBody += (i < sorting.length - 1) ? "{" : "return 0;";
            }
            functionBody += Array(sorting.length).join("}") + " return -1;";
            var f = new Function('a', 'b', functionBody);
            return arr.sort(f);
        };
        IndexedDbWrapper.prototype.paging = function (arr, query) {
            var startIndex = 0, endIndex = 0;
            // When pageIndex is -1 then don't check for pages and load all rows.
            if (query.pageIndex >= 0) {
                startIndex = query.pageIndex * query.pageSize;
                endIndex = ((query.pageIndex + 1) * query.pageSize);
            }
            else if (query.pageSize !== -1) {
                endIndex = query.pageSize; // top statement
            }
            return arr.slice(startIndex, endIndex);
        };
        IndexedDbWrapper.prototype.getObjectStore = function (objId, mode) {
            var db = this.db;
            if (db && !this.isDisposed) {
                var transaction = db.transaction([objId], mode);
                return transaction.objectStore(objId);
            }
            return null;
        };
        IndexedDbWrapper.prototype.insertBookmark = function (obj, bookmark, markForOnlineSync) {
            var _this = this;
            var deferred = jQuery.Deferred();
            if (this.isDisposed) {
                return deferred.resolve().promise();
            }
            var identityObjectStore = this.getObjectStore(IndexedDbWrapper.identityTableName, "readwrite");
            // Always insert something in queue, this way we can remove them before a synchronization
            // If the localsync didn't return them they didn't belong there anyway
            var countRequest = identityObjectStore.count();
            countRequest.onsuccess = function (event) {
                var objObjectStore = _this.getObjectStore(obj.objId, "readwrite");
                var count = countRequest.result;
                var objProps = bookmark.obj.objProps.array;
                for (var i = 0; i < objProps.length; i++) {
                    if (objProps[i].identity) {
                        var replacer = new IdentityReplacer(++count);
                        identityObjectStore.add(replacer);
                        bookmark.addCondition(new TSFBookmarkCondition(objProps[i], replacer.id));
                    }
                }
                var bookmarkObject = bookmark.toObject();
                IndexedDb.addIndexes(obj, bookmarkObject);
                var addReq = objObjectStore.add(bookmarkObject);
                addReq.onsuccess = function (event) {
                    // Wait for the add to complete
                    objObjectStore.transaction.oncomplete = function () {
                        _this.storeQueueItem(new QueueItem((markForOnlineSync ? 0 /* Insert */ : 4 /* VolatileInsert */), bookmark))
                            .done(function () {
                            deferred.resolve(new IndexedDb.IndexedDbArrayReader([bookmarkObject]));
                        })
                            .fail(function (message) {
                            deferred.reject(message);
                        });
                    };
                };
                addReq.onerror = function (event) {
                    deferred.reject("IndexedDb error: couldn't insert the row in the object store. " + event.message);
                };
            };
            countRequest.onerror = function (event) {
                deferred.reject("IndexedDb error: couldn't count the identity object store. " + event.message);
            };
            return deferred.promise();
        };
        IndexedDbWrapper.prototype.storeQueueItem = function (queueItem) {
            var _this = this;
            var deferred = jQuery.Deferred();
            var queueObjectStore = this.getObjectStore(IndexedDbWrapper.queueTableName, "readwrite");
            if (queueObjectStore) {
                var addReq = queueObjectStore.add(queueItem);
                addReq.onsuccess = function () {
                    // Wait for the transaction to complete or else the data might not be uptodate when a refresh is called during the transaction
                    queueObjectStore.transaction.oncomplete = function () {
                        _this.application.serverSync.newOfflineChangeAdded();
                        deferred.resolve();
                    };
                };
                addReq.onerror = function (event) {
                    deferred.reject("IndexedDb error: couldn't store the queue item. " + event.message);
                };
            }
            else {
                deferred.resolve();
            }
            return deferred.promise();
        };
        IndexedDbWrapper.prototype.updateRow = function (obj, oldPkBookmark, bookmark, markForOnlineSync) {
            var _this = this;
            var deferred = jQuery.Deferred();
            var objectStore = this.getObjectStore(obj.objId, "readwrite");
            if (!objectStore) {
                return deferred.reject("Store not found").promise();
            }
            var index = objectStore.index(IndexedDbWrapper.pkIndexName);
            index.openCursor(IndexedDb.createPKKeyRangeFromBookmark(obj, oldPkBookmark)).onsuccess = function (event) {
                var cursor = event.target.result;
                if (cursor) {
                    try {
                        var bookmarkObject_1 = bookmark.toObject();
                        bookmarkObject_1["tsf_id"] = cursor.primaryKey;
                        IndexedDb.addIndexes(obj, bookmarkObject_1);
                        var request = cursor.update(bookmarkObject_1);
                        request.onsuccess = function (event) {
                            // Wait for the transaction to complete or else the data might not be uptodate when a refresh is called during the transaction
                            objectStore.transaction.oncomplete = function () {
                                if (markForOnlineSync) {
                                    _this.storeQueueItem(new QueueItem(1 /* Update */, bookmark, oldPkBookmark))
                                        .done(function () {
                                        deferred.resolve(new IndexedDb.IndexedDbArrayReader([bookmarkObject_1]));
                                    })
                                        .fail(function (message) {
                                        deferred.reject(message);
                                    });
                                }
                                else {
                                    deferred.resolve(new IndexedDb.IndexedDbArrayReader([bookmarkObject_1]));
                                }
                            };
                        };
                        request.onerror = function (event) {
                            deferred.reject("IndexedDb error: " + event.message);
                        };
                    }
                    catch (err) {
                        deferred.reject("IndexedDb error: " + err);
                    }
                }
                else {
                    deferred.reject("Row not found.");
                }
            };
            return deferred.promise();
        };
        IndexedDbWrapper.prototype.deleteRow = function (obj, pkBookmark, markForOnlineSync) {
            var _this = this;
            var deferred = jQuery.Deferred();
            var objectStore = this.getObjectStore(obj.objId, "readwrite");
            if (!objectStore) {
                return deferred.reject("Store not found").promise();
            }
            var index = objectStore.index(IndexedDbWrapper.pkIndexName);
            index.openCursor(IndexedDb.createPKKeyRangeFromBookmark(obj, pkBookmark)).onsuccess = function (event) {
                var cursor = event.target.result;
                if (cursor) {
                    var request = cursor['delete']();
                    request.onsuccess = function (event) {
                        objectStore.transaction.oncomplete = function () {
                            if (markForOnlineSync) {
                                _this.storeQueueItem(new QueueItem(2 /* Delete */, pkBookmark))
                                    .done(function () {
                                    deferred.resolve(new IndexedDb.IndexedDbArrayReader([pkBookmark]));
                                })
                                    .fail(function (message) {
                                    deferred.reject(message);
                                });
                            }
                            else {
                                deferred.resolve(new IndexedDb.IndexedDbArrayReader([pkBookmark]));
                            }
                        };
                    };
                    request.onerror = function (event) {
                        deferred.reject("IndexedDb error: " + event.message);
                    };
                }
                else {
                    // No more entries
                    deferred.reject("row not found.");
                }
            };
            return deferred.promise();
        };
        IndexedDbWrapper.prototype.queueTask = function (row) {
            var queueItem = new QueueItem(3 /* Task */, row);
            return this.storeQueueItem(queueItem);
        };
        IndexedDbWrapper.prototype.clearQueue = function () {
            var resolveList = [];
            resolveList.push(this.clearStore(IndexedDb.IndexedDbWrapper.queueTableName));
            resolveList.push(this.clearStore(IndexedDb.IndexedDbWrapper.identityTableName));
            return jQuery.when.apply(undefined, resolveList);
        };
        IndexedDbWrapper.prototype.clearOfflineData = function () {
            var objs = this.application.model.objs.array;
            var resolveList = [];
            for (var i = 0; i < objs.length; i++) {
                resolveList.push(this.clearStore(objs[i].objId));
            }
            return jQuery.when.apply(undefined, resolveList);
        };
        IndexedDbWrapper.prototype.clearResponseBufferStore = function () {
            return this.clearStore(IndexedDb.IndexedDbWrapper.responseBufferTableName);
        };
        IndexedDbWrapper.prototype.clearStore = function (storeId) {
            var deferred = jQuery.Deferred();
            var db = this.db;
            if (!db)
                return deferred.resolve().promise();
            try {
                var transaction = db.transaction([storeId], "readwrite");
                var objectStore = transaction.objectStore(storeId);
                var req = objectStore.clear();
                req.onsuccess = function () {
                    deferred.resolve();
                };
                req.onerror = function () {
                    deferred.reject();
                };
            }
            catch (ex) {
                deferred.reject();
            }
            return deferred.promise();
        };
        IndexedDbWrapper.prototype.resolveIdentitiesInRow = function (objView, bookmark) {
            var _this = this;
            var deferred = jQuery.Deferred(), resolveList = [], newBookmark = {};
            jQuery.extend(true, newBookmark, bookmark);
            for (var i = 0; i < newBookmark.conditions.length; i++) {
                var condition = newBookmark.conditions[i];
                var objViewProp = objView.objViewProps.object[condition.objProp.objPropId];
                if (objViewProp.objProp.identity || (objViewProp.hasLookup() && objViewProp.lookup.dataProp.objProp.identity)) {
                    if (typeof condition.value === "string" && condition.value.indexOf(IdentityReplacer.prefix) > -1) {
                        resolveList.push((function (condition) {
                            return _this.application.serverSync.resolveIdentity(condition.value).then(function (databaseId) {
                                condition.value = databaseId;
                            });
                        })(condition));
                    }
                }
            }
            jQuery.when.apply(undefined, resolveList).always(function () { deferred.resolve(newBookmark); });
            return deferred.promise();
        };
        IndexedDbWrapper.prototype.dispose = function () {
            if (this.db) {
                this.db.close();
                this.isDisposed = true;
            }
        };
        IndexedDbWrapper.responseBufferTableName = "tsf_mobile_response_buffer";
        IndexedDbWrapper.syncGroupLastSyncTableName = "tsf_mobile_sync_group_last_sync";
        IndexedDbWrapper.queueTableName = "tsf_mobile_queue";
        IndexedDbWrapper.identityTableName = "tsf_mobile_identity";
        IndexedDbWrapper.fileTableName = "tsf_mobile_files";
        IndexedDbWrapper.fileRefTableName = "tsf_mobile_file_ref";
        IndexedDbWrapper.pkIndexName = "tsf_pk";
        IndexedDbWrapper.sortIndexName = "tsf_sort";
        IndexedDbWrapper.lfTrnslPrefix = "tsf_lf_trnsl_"; //lookup translation
        IndexedDbWrapper.lfPrefix = "tsf_lf_"; //lookupfilter
        IndexedDbWrapper.olPrefix = "tsf_ol_"; //objlink
        return IndexedDbWrapper;
    }());
    IndexedDb.IndexedDbWrapper = IndexedDbWrapper;
    var FilterHelper;
    (function (FilterHelper) {
        var Where = /** @class */ (function () {
            function Where(whereClause, name, ignoreCasing, operator) {
                this.whereClause = whereClause;
                Where.createPrefilterFunctions();
                for (var i in Where.knownFilters) {
                    var metaData = {
                        propertyName: name,
                        operator: operator,
                        ignoreCasing: ignoreCasing,
                        isValid: Where.knownFilters[i].isValid
                    };
                    this[i] = Where.knownFilters[i].callback(metaData, whereClause);
                }
            }
            Where.supportsFilterCondition = function (condition) {
                Where.createPrefilterFunctions();
                return Where.knownFilters[condition] !== undefined;
            };
            Where.createPrefilterFunctions = function () {
                if (!Where.knownFilters) {
                    Where.knownFilters = {};
                    Where.createFilter("equalTo", function (data, meta) {
                        var value = data[meta.propertyName];
                        // Only do something with casing if both values are not undefined.
                        if (meta.ignoreCasing && value && meta.value) {
                            return ("" + value).toLowerCase() == ("" + meta.value).toLowerCase();
                        }
                        return ("" + value) == ("" + meta.value);
                    }, function (data, whereClause) {
                        return function (value) {
                            whereClause.filters.push(data);
                            data.value = value;
                            return whereClause;
                        };
                    });
                    Where.createFilterCopy("notEqualTo", "equalTo", function (data, whereClause) {
                        data.isNotClause = true;
                    });
                    Where.createFilter("dateEqualTo", function (data, meta) {
                        var value = data[meta.propertyName];
                        if (value && meta.value) {
                            return ("" + value) == ("" + meta.value);
                        }
                        return false;
                    }, function (data, whereClause) {
                        return function (value) {
                            whereClause.filters.push(data);
                            data.value = (typeof value == "string" ? new Date(value) : value);
                            return whereClause;
                        };
                    });
                    Where.createFilterCopy("dateNotEqualTo", "dateEqualTo", function (data, whereClause) {
                        data.isNotClause = true;
                    });
                    Where.createFilter("between", function (data, meta) {
                        var value = data[meta.propertyName];
                        return (value > meta.value && value < meta.untilValue);
                    }, function (data, whereClause) {
                        return function (value, untilValue) {
                            whereClause.filters.push(data);
                            data.value = value;
                            data.untilValue = untilValue;
                            return whereClause;
                        };
                    });
                    Where.createFilterCopy("notBetween", "between", function (data, whereClause) {
                        data.isNotClause = true;
                    });
                    Where.createFilter("greaterThan", function (data, meta) {
                        var value = data[meta.propertyName];
                        return (value > meta.value || (meta.includeValues && value == meta.value));
                    }, function (data, whereClause) {
                        return function (value) {
                            whereClause.filters.push(data);
                            data.value = value;
                            return whereClause;
                        };
                    });
                    Where.createFilterCopy("greaterThanOrEqualTo", "greaterThan", function (data, whereClause) {
                        data.includeValues = true;
                    });
                    Where.createFilter("smallerThan", function (data, meta) {
                        var value = data[meta.propertyName];
                        return (value < meta.value || (meta.includeValues && value == meta.value));
                    }, function (data, whereClause) {
                        return function (value) {
                            whereClause.filters.push(data);
                            data.value = value;
                            return whereClause;
                        };
                    });
                    Where.createFilterCopy("smallerThanOrEqualTo", "smallerThan", function (data, whereClause) {
                        data.includeValues = true;
                    });
                    Where.createFilter("contains", function (data, meta) {
                        var value = data[meta.propertyName] + "";
                        var filterValue = meta.value ? (meta.value + "").toLowerCase() : null;
                        return value.toLowerCase().indexOf(filterValue) >= 0;
                    }, function (data, whereClause) {
                        return function (value) {
                            whereClause.filters.push(data);
                            data.value = (value + "").toLowerCase();
                            return whereClause;
                        };
                    });
                    Where.createFilterCopy("doesNotContain", "contains", function (data, whereClause) {
                        data.isNotClause = true;
                    });
                    Where.createFilter("isNull", function (data, meta) {
                        var value = data[meta.propertyName];
                        return (value === undefined || value === null);
                    }, function (data, whereClause) {
                        return function () {
                            whereClause.filters.push(data);
                            return whereClause;
                        };
                    });
                    Where.createFilterCopy("isNotNull", "isNull", function (data, whereClause) {
                        data.isNotClause = true;
                    });
                    Where.createFilter("startsWith", function (data, meta) {
                        var value = data[meta.propertyName];
                        //http://jsperf.com/js-startswith/6, substring seems to perform really well
                        var success = meta.value;
                        return (value != null && (value + "").substring(0, success.length).toLowerCase() === success);
                    }, function (data, whereClause) {
                        return function (value) {
                            whereClause.filters.push(data);
                            data.value = (value + "").toLowerCase();
                            return whereClause;
                        };
                    });
                    Where.createFilterCopy("doesNotStartWith", "startsWith", function (data, whereClause) {
                        data.isNotClause = true;
                    });
                }
            };
            Where.createFilter = function (name, isValid, callback) {
                Where.knownFilters[name] = { name: name, isValid: isValid, callback: callback };
            };
            Where.createFilterCopy = function (name, basedOnFilter, callback) {
                var copy = Where.knownFilters[basedOnFilter];
                if (copy) {
                    var filter = {
                        name: name,
                        isValid: copy.isValid,
                        callback: function (filterdata, whereClause) {
                            return function () {
                                var args = [];
                                for (var _i = 0; _i < arguments.length; _i++) {
                                    args[_i] = arguments[_i];
                                }
                                callback(filterdata, whereClause);
                                // Use apply instead of call because the ammount of not null args is unknown.
                                return copy.callback(filterdata, whereClause).apply(null, args);
                            };
                        }
                    };
                    Where.knownFilters[name] = filter;
                }
            };
            return Where;
        }());
        FilterHelper.Where = Where;
        var WhereClauseHelper = /** @class */ (function () {
            function WhereClauseHelper() {
                this.filters = [];
            }
            WhereClauseHelper.prototype.where = function (name, ignoreCasing, operator) {
                return new Where(this, name, ignoreCasing, operator);
            };
            WhereClauseHelper.prototype.isRowValid = function (row) {
                var allValid = true;
                var oneOrValid = false;
                var containsOrOperators = false;
                for (var i = 0; i < this.filters.length; i++) {
                    var filterMeta = this.filters[i];
                    var valid = filterMeta.isValid(row, filterMeta);
                    if (filterMeta.isNotClause)
                        valid = !valid;
                    // Instantly return when an AND filter doesn't apply or when an OR filter does apply.
                    // Note that this code doesn't support mixed use of AND and OR operators.
                    if (filterMeta.operator === 0 /* And */ && !valid) {
                        return false;
                    }
                    if (filterMeta.operator === 1 /* Or */) {
                        containsOrOperators = true;
                        if (valid) {
                            oneOrValid = true;
                        }
                    }
                }
                if (containsOrOperators) {
                    return allValid && oneOrValid;
                }
                return allValid;
            };
            WhereClauseHelper.prototype.filterStore = function (store, query, objView, wrapper) {
                var _this = this;
                var deferred = jQuery.Deferred();
                if (query.whereClause.combinedFilterConditions.length > 0 && objView !== null) {
                    this.handleLookupFilters(query, objView, wrapper).then(function (modifiedQuery) {
                        _this.filters = [];
                        IndexedDbWrapper.createFilterHelperInner(_this, query.whereClause.contextFilters);
                        IndexedDbWrapper.createFilterHelperInner(_this, modifiedQuery.whereClause.combinedFilterConditions, 1 /* Or */);
                        IndexedDbWrapper.createFilterHelperPrefilters(_this, query.whereClause.prefilterState);
                        var objId = objView.obj.objId;
                        var _a = _this.createRequest(wrapper.getObjectStore(objId, "readonly"), query), request = _a.request, filterPaged = _a.filterPaged;
                        _this.fetchFiltered(filterPaged, request, deferred, query);
                    });
                }
                else {
                    var _a = this.createRequest(store, query), request = _a.request, filterPaged = _a.filterPaged;
                    this.fetchFiltered(filterPaged, request, deferred, query);
                }
                return deferred.promise();
            };
            WhereClauseHelper.prototype.createRequest = function (store, query) {
                var filterPaged = false;
                var request;
                // Use the index and keyRange if there is one specified.
                if (query.preferredIndex) {
                    var sorting = "next";
                    if (query.preferredIndex === IndexedDb.IndexedDbWrapper.sortIndexName &&
                        query.currentSorting && query.currentSorting.length > 0 &&
                        query.currentSorting[0].direction == Enum.SortDirection.Descending) {
                        sorting = "prev";
                    }
                    var index = store.index(query.preferredIndex);
                    if (query.keyRange) {
                        filterPaged = query.indexSorted;
                        request = index.openCursor(query.keyRange, sorting);
                    }
                    else {
                        request = index.openCursor(null, sorting);
                        filterPaged = true;
                    }
                }
                else {
                    request = store.openCursor();
                }
                return { request: request, filterPaged: filterPaged };
            };
            WhereClauseHelper.prototype.fetchFiltered = function (filterPaged, request, deferred, query) {
                // Filter paged when there is a preferredIndex but no keyRange.
                // We only use this for cases when we don't need to filter but when we do need sorting
                // Example, master subjects or lookups without a filter.
                if (filterPaged) {
                    this.filterPaged(request, deferred, query);
                }
                else {
                    this.filterUnPaged(request, deferred);
                }
            };
            WhereClauseHelper.prototype.handleLookupFilters = function (query, objView, wrapper) {
                var combFilterConditions = query.whereClause.combinedFilterConditions;
                var promises = [];
                var _loop_1 = function (i) {
                    var combinedFilterCondition = combFilterConditions[i];
                    var viewProp = objView.objViewProps.array.filter(function (p) {
                        return p.objProp.objPropId == combinedFilterCondition.targetPropId;
                    })[0];
                    if (viewProp.isTranslatedLookup()) {
                        //Prevent already handled lookup conditions from being handled 2 times
                        //(this happens when the controller gets reused by going to menu and back)
                        if (combinedFilterCondition.comparision !== Enum.FilterCondition.Contains) {
                            return "break";
                        }
                        if (viewProp.lookup.targetObjView.offlineAvailable) {
                            promises.push(this_1.getFilterConditionsFromLookup(combinedFilterCondition, viewProp.lookup, wrapper));
                        }
                        combFilterConditions.splice(i, 1);
                    }
                };
                var this_1 = this;
                for (var i = combFilterConditions.length - 1; i >= 0; i--) {
                    var state_1 = _loop_1(i);
                    if (state_1 === "break")
                        break;
                }
                if (promises.length === 0) {
                    var deferred = jQuery.Deferred();
                    deferred.resolve(query);
                    return deferred.promise();
                }
                return Utils.whenAll(promises).then(function () {
                    var args = [];
                    for (var _i = 0; _i < arguments.length; _i++) {
                        args[_i] = arguments[_i];
                    }
                    args.forEach(function (addedCombFilterConditions) {
                        query.whereClause.combinedFilterConditions = query.whereClause.combinedFilterConditions.concat(addedCombFilterConditions);
                    });
                    return query;
                });
            };
            WhereClauseHelper.prototype.getFilterConditionsFromLookup = function (filter, lookup, wrapper) {
                var results = [];
                var deferred = jQuery.Deferred();
                var lookupObj = lookup.targetObjView.objId;
                var lookupDisplayProp = lookup.listProp.objPropId;
                var objectStore = wrapper.getObjectStore(lookupObj, "readonly");
                var request = objectStore.openCursor();
                var filterHelper = new WhereClauseHelper();
                filterHelper.where(lookupDisplayProp, true, 0 /* And */)["contains"](filter.dbFilterValue);
                request.onsuccess = function (event) {
                    var cursor = event.target.result;
                    if (cursor) {
                        var fh = filterHelper;
                        if (filterHelper.isRowValid(cursor.value)) {
                            results.push(new FilterContractCondition(filter.targetProp, Enum.FilterCondition.EqualTo, cursor.value[lookup.dataProp.objPropId]));
                        }
                        cursor.continue();
                    }
                    else {
                        deferred.resolve(results);
                    }
                };
                return deferred;
            };
            WhereClauseHelper.prototype.filterUnPaged = function (request, deferred) {
                var result = [], that = this;
                request.onsuccess = function (event) {
                    var cursor = event.target.result;
                    if (cursor) {
                        if (that.isRowValid(cursor.value)) {
                            result.push(cursor.value);
                        }
                        cursor['continue']();
                    }
                    else {
                        // no more records
                        deferred.resolve(result);
                    }
                };
            };
            WhereClauseHelper.prototype.filterPaged = function (request, deferred, query) {
                var result = [], that = this, pageIndex = query.usePagination ? query.pageIndex : -1, pageSize = query.usePagination ? query.pageSize : -1;
                var maxRows = -1, foundRowIndex = 0, currentRows = 0;
                // When pageIndex is -1 then don't check for pages and load all rows.
                if (pageIndex >= 0) {
                    currentRows = pageIndex * pageSize;
                    maxRows = ((pageIndex + 1) * pageSize);
                }
                else if (pageSize !== -1) {
                    maxRows = pageSize; // top statement
                }
                request.onsuccess = function (event) {
                    var cursor = event.target.result;
                    if (cursor) {
                        // Only continue to add new rows when the amount of found rows is smaller then the maximum rows minus the current amount of rows.
                        if (maxRows === -1 || result.length < (maxRows - currentRows)) {
                            if (that.isRowValid(cursor.value)) {
                                foundRowIndex++;
                                // Only add the row to the result when paging is disabled or when the foundRowIndex is greather than the currrent amount of rows.
                                if (pageIndex === -1 || currentRows < foundRowIndex) {
                                    result.push(cursor.value);
                                }
                            }
                            cursor['continue']();
                        }
                        else {
                            deferred.resolve(result);
                            return;
                        }
                    }
                    else {
                        // No more records.
                        deferred.resolve(result);
                    }
                };
            };
            return WhereClauseHelper;
        }());
        FilterHelper.WhereClauseHelper = WhereClauseHelper;
    })(FilterHelper || (FilterHelper = {}));
})(IndexedDb || (IndexedDb = {}));
