var IndexedDb;
(function (IndexedDb) {
    var IndexedDbArrayReader = /** @class */ (function () {
        function IndexedDbArrayReader(itemList) {
            this.itemList = itemList;
        }
        Object.defineProperty(IndexedDbArrayReader.prototype, "length", {
            get: function () {
                return this.itemList.length;
            },
            enumerable: true,
            configurable: true
        });
        IndexedDbArrayReader.prototype.item = function (index) {
            var i = this.itemList[index];
            if (i)
                return new IndexedDbDataReader(i);
            return null;
        };
        IndexedDbArrayReader.prototype.importData = function (reader) {
            for (var i = 0; i < reader.length; i++) {
                var item = reader.item(i);
                if (item) {
                    this.itemList.push(item.itemObject);
                }
            }
        };
        IndexedDbArrayReader.prototype.join = function (reader, selectParts, joinConditions) {
            // Check for every row in this reader if they match with a row in the other reader.
            for (var i = 0; i < this.itemList.length; i++) {
                var sourceItem = this.item(i);
                for (var g = 0; g < reader.length; g++) {
                    var targetItem = reader.item(g);
                    // Check if the join conditions match. Skip the row if they don't.
                    if (Utils.evaluateJoinConditions(joinConditions, sourceItem, targetItem)) {
                        // Actually join the readers, only copy the given select parts.
                        for (var h = 0; h < selectParts.length; h++) {
                            var readerObjPropValue = targetItem.getValueForObjProp(selectParts[h].objProp);
                            sourceItem.setValue(selectParts[h].aliasWebServiceId, readerObjPropValue);
                        }
                    }
                }
            }
        };
        IndexedDbArrayReader.prototype.serialize = function () {
            return Utils.serializeJSON(this.itemList);
        };
        IndexedDbArrayReader.deserialize = function (json) {
            return new IndexedDbArrayReader(Utils.deserializeJSON(json));
        };
        return IndexedDbArrayReader;
    }());
    IndexedDb.IndexedDbArrayReader = IndexedDbArrayReader;
    var IndexedDbDataReader = /** @class */ (function () {
        function IndexedDbDataReader(itemObject) {
            this.itemObject = itemObject;
        }
        IndexedDbDataReader.prototype.item = function (name) {
            var val = this.getItem(name, this.itemObject);
            if (val)
                return new IndexedDbDataReader(val);
            return null;
        };
        IndexedDbDataReader.prototype.list = function (name) {
            var list = this.itemObject[name];
            if (list) {
                return new IndexedDbArrayReader(Object.prototype.toString.call(list) == "[object Array]" ? list : [list]);
            }
            return null;
        };
        IndexedDbDataReader.prototype.getValueForObjProp = function (objProp, defaultValue) {
            if (defaultValue === void 0) { defaultValue = null; }
            if (objProp.isFileColumn) {
                var file = this.getFile(objProp.objPropId);
                if (file)
                    return file;
            }
            else if (objProp.xmlDttp == "date") {
                // Also for sqltime datatype the SF returns date; so this  is actually a hack
                // because we can't really now if they have a control_id time with a time or a datetime datatype on the database
                // TCP 49459
                var d = this.getString(objProp.objPropId);
                if (d) {
                    var tmp = moment(d);
                    if (!tmp.isValid() && objProp.controlId == Enum.ControlType.TIME) {
                        // try time as a fallback
                        return moment(d, "hh:mm:ss.sss").toDate();
                    }
                    return tmp.toDate();
                }
            }
            else if (objProp.isNumeric()) {
                var value = this.itemObject[objProp.objPropId];
                if (typeof value === "string" && value.indexOf(IdentityReplacer.prefix) === 0) {
                    return value;
                }
                return this.getNumber(objProp.objPropId, defaultValue);
            }
            else {
                return this.getString(objProp.objPropId, defaultValue);
            }
            return defaultValue;
        };
        IndexedDbDataReader.prototype.getFile = function (name) {
            var fileItem = this.getItem(name, this.itemObject);
            if (fileItem) {
                var fileContract = void 0;
                var filePath = fileItem.getString("FilePath");
                if (!Utils.isEmpty(filePath)) {
                    fileContract = FileContract.CreateFromFilePath(filePath, undefined, undefined, undefined);
                    fileContract.File = fileItem.getString("File");
                    fileContract.Thumbnail = fileItem.getString("Thumbnail");
                    fileContract.HashCode = fileItem.getNumber("HashCode");
                    return fileContract;
                }
            }
            return null;
        };
        IndexedDbDataReader.prototype.setFile = function (name, fileContract) {
            var fileItem = this.itemObject[name];
            if (!fileItem) {
                this.itemObject[name] = fileItem = {};
            }
            if (fileContract && !Utils.isEmpty(fileContract.FileName)) {
                if (fileContract.FilePath != null) {
                    fileItem["FilePath"] = fileContract.FilePath;
                }
                if (fileContract.FileName != null) {
                    fileItem["FileName"] = fileContract.FileName;
                }
                if (fileContract.File != null) {
                    fileItem["File"] = fileContract.File;
                }
                if (fileContract.Thumbnail != null) {
                    fileItem["Thumbnail"] = fileContract.Thumbnail;
                }
                if (fileContract.HashCode != null) {
                    fileItem["HashCode"] = fileContract.HashCode;
                }
            }
            else {
                this.itemObject[name] = null;
            }
        };
        IndexedDbDataReader.prototype.setValue = function (name, value) {
            this.itemObject[name] = value;
        };
        IndexedDbDataReader.prototype.getString = function (name, defaultValue) {
            if (defaultValue === void 0) { defaultValue = null; }
            var value = this.itemObject[name];
            // With offline defaults we sometimes put a boolean in here, when it is false it returns defaultValue instead of false
            if (value !== undefined && value !== null) {
                return value;
            }
            return defaultValue;
        };
        IndexedDbDataReader.prototype.getBool = function (name, defaultValue) {
            if (defaultValue === void 0) { defaultValue = false; }
            var value = this.itemObject[name];
            return Utils.convertToBoolean(value, defaultValue);
        };
        IndexedDbDataReader.prototype.getNumber = function (name, defaultValue) {
            if (defaultValue === void 0) { defaultValue = 0; }
            var value = this.itemObject[name];
            if (value !== undefined && value !== null) {
                return +value;
            }
            return defaultValue;
        };
        IndexedDbDataReader.prototype.getNullableNumber = function (name) {
            return this.getNumber(name, null);
        };
        IndexedDbDataReader.prototype.getNullableBool = function (name) {
            return this.getBool(name, null);
        };
        IndexedDbDataReader.prototype.getItem = function (key, object) {
            if (Object.prototype.toString.call(object) == "[object Object]") {
                var keys = Object.keys(object);
                for (var i = 0; i < keys.length; i++) {
                    if (keys[i] == key) {
                        if (Object.prototype.toString.call(object[key]) == "[object Object]") {
                            return new IndexedDbDataReader(object[key]);
                        }
                        else {
                            return object[key];
                        }
                    }
                    else {
                        var ret = this.getItem(key, object[keys[i]]);
                        if (ret) {
                            return ret;
                        }
                    }
                }
            }
            return null;
        };
        return IndexedDbDataReader;
    }());
    IndexedDb.IndexedDbDataReader = IndexedDbDataReader;
})(IndexedDb || (IndexedDb = {}));
