var FileCache = /** @class */ (function () {
    function FileCache(application) {
        this.application = application;
    }
    FileCache.prototype.openStore = function (store, mode) {
        var deferred = jQuery.Deferred();
        var database = this.application.getDatabase();
        if (!database || this.application.databaseIsDisposed()) {
            return deferred.reject().promise();
        }
        var transaction = database.transaction(store, mode);
        return deferred.resolve(transaction.objectStore(store)).promise();
    };
    FileCache.prototype.downloadFile = function (obj, prop, row) {
        var deferred = jQuery.Deferred();
        var filterContract = FilterContract.fromDataRow(obj, row);
        var requestBuilder = RequestBuilderFactory.createOnlineRequestBuilder(this.application, obj.defaultObjView)
            .addSelectionContract([prop.objPropId])
            .addFilterContract(filterContract)
            .addFileOptionsContract({
            IncludeHashCodes: false,
            IncludeFiles: true,
            IncludeThumbnails: false
        });
        requestBuilder.setCallbackMethod(function (reader, status) {
            if (reader && reader.length == 1 && status.success) {
                var value = reader.item(0).getValueForObjProp(prop);
                if (value && value.File) {
                    deferred.resolve(value.File);
                    return;
                }
            }
            deferred.resolve(null);
        });
        requestBuilder.get();
        return deferred.promise();
    };
    FileCache.prototype.removeFileReference = function (fileContract) {
        var _this = this;
        var deferred = jQuery.Deferred();
        if (fileContract.HashCode == null) {
            // A file can be removed without having an HashCode.
            // This occurs when for instance a file is online added and removed before a LocalSync was completed.
            deferred.resolve();
        }
        else {
            this.openStore(IndexedDb.IndexedDbWrapper.fileRefTableName, "readwrite").then(function (refStore) {
                var refReq = refStore.openCursor(IDBKeyRange.only(fileContract.HashCode));
                refReq.onsuccess = function () {
                    var refCursor = refReq.result;
                    if (refCursor) {
                        // Remove the file from the cache when the RefCount is 1 or lower.
                        if (refCursor.value.RefCount <= 1) {
                            refCursor['delete']();
                            _this.removeFileFromCache(fileContract.HashCode, deferred);
                            return;
                        }
                        else {
                            // Subtract 1 from the RefCount and update the value.
                            var value = refCursor.value;
                            value.RefCount--;
                            refCursor.update(value);
                        }
                    }
                    deferred.resolve();
                };
            }, function () {
                deferred.reject();
            });
        }
        return deferred.promise();
    };
    FileCache.prototype.removeFileFromCache = function (hashCode, deferred) {
        this.openStore(IndexedDb.IndexedDbWrapper.fileTableName, "readwrite").then(function (fileStore) {
            var fileReq = fileStore.openCursor(IDBKeyRange.only(hashCode));
            fileReq.onsuccess = function () {
                var fileCursor = fileReq.result;
                if (fileCursor) {
                    fileCursor['delete']();
                }
            };
            fileStore.transaction.oncomplete = fileStore.transaction.onerror = function () {
                deferred.resolve();
            };
        }, function () {
            deferred.reject();
        });
        return deferred.promise();
    };
    FileCache.prototype.updateFileCache = function (obj, objProp, newRow) {
        var _this = this;
        var deferred = jQuery.Deferred();
        var database = this.application.getDatabase(), transaction = database.transaction(obj.objId, "readonly"), store = transaction.objectStore(obj.objId);
        var index = store.index(IndexedDb.IndexedDbWrapper.pkIndexName);
        var req = index.openCursor(IndexedDb.createPKKeyRangeFromBookmark(obj, TSFPkBookmark.fromObject(obj, newRow)));
        req.onsuccess = function () {
            var cursor = req.result;
            if (cursor) {
                var oldRow = cursor.value;
                var existingFile = oldRow[objProp.objPropId];
                var newFile = newRow[objProp.objPropId];
                if (!existingFile && newFile) {
                    // File is added
                    _this.addFileReference(obj, objProp, newRow);
                }
                else if (existingFile && !newFile) {
                    // File is removed
                    _this.removeFileReference(existingFile);
                }
                else if (existingFile && newFile && existingFile.HashCode != newFile.HashCode) {
                    // File is edited
                    _this.removeFileReference(existingFile);
                    _this.addFileReference(obj, objProp, newRow);
                }
                deferred.resolve();
            }
            else {
                if (newRow[objProp.objPropId]) {
                    _this.addFileReference(obj, objProp, newRow);
                }
                deferred.resolve();
            }
        };
        return deferred.promise();
    };
    FileCache.prototype.addFileReference = function (obj, objProp, newRow) {
        var _this = this;
        var deferred = jQuery.Deferred();
        var newFileReference = newRow[objProp.objPropId];
        this.openStore(IndexedDb.IndexedDbWrapper.fileTableName, "readwrite").then(function (fileStore) {
            // Open the filestore with the same HashCode
            var fileReq = fileStore.openCursor(IDBKeyRange.only(newFileReference.HashCode));
            fileReq.onsuccess = function () {
                // Download and add the file when no file with the same HashCode exists.
                var fileCursor = fileReq.result;
                if (!fileCursor) {
                    _this.downloadFile(obj, objProp, newRow).then(function (base64File) {
                        // Open a new transaction because it might take a while to download the file.
                        _this.openStore(IndexedDb.IndexedDbWrapper.fileTableName, "readwrite").then(function (addStore) {
                            // Copy the object because we don't want the files in the offline dataset.
                            var fileContract = jQuery.extend({}, newFileReference);
                            fileContract.File = base64File;
                            addStore.add(fileContract);
                        }, function (err) {
                            console.error("Couldn't download and add the file.", err);
                            deferred.reject();
                        });
                    });
                }
                // Open the file reference store with the same HashCode.
                _this.openStore(IndexedDb.IndexedDbWrapper.fileRefTableName, "readwrite").then(function (refStore) {
                    var refReq = refStore.openCursor(IDBKeyRange.only(newFileReference.HashCode));
                    refReq.onsuccess = function () {
                        // Bump the RefCount when the table contains a reference to the HashCode.
                        // Otherwise add a new reference with RefCount 1.
                        var refCursor = refReq.result;
                        if (refCursor) {
                            var value = refCursor.value;
                            value.RefCount++;
                            refCursor.update(value);
                        }
                        else {
                            refStore.add({ RefCount: 1, HashCode: newFileReference.HashCode });
                        }
                    };
                    refStore.transaction.oncomplete = refStore.transaction.onerror = function () {
                        deferred.resolve();
                    };
                }, function (err) {
                    console.error("Couldn't open file from store", err);
                    deferred.reject();
                });
            };
        }, function () {
            deferred.reject();
        });
        return deferred.promise();
    };
    FileCache.prototype.getCachedFile = function (hashCode) {
        var deferred = jQuery.Deferred();
        this.openStore(IndexedDb.IndexedDbWrapper.fileTableName, "readonly").always(function (store) {
            var req = store.openCursor(IDBKeyRange.only(hashCode));
            req.onsuccess = function () {
                var cursor = req.result;
                if (cursor) {
                    var file = cursor.value;
                    deferred.resolve(file.File);
                }
                else {
                    deferred.resolve(null);
                }
            };
        });
        return deferred.promise();
    };
    return FileCache;
}());
