/**
 *
 */
var SyncGroup = /** @class */ (function () {
    function SyncGroup(syncGroupId, syncInterval) {
        this.syncGroupId = syncGroupId;
        this.syncInterval = syncInterval;
        this.atomic = false;
        this.objs = new Collections.List();
    }
    return SyncGroup;
}());
var IntervalToken = /** @class */ (function () {
    /**
     * @param syncInterval in minutes
     */
    function IntervalToken(syncInterval) {
        this.syncInterval = syncInterval;
    }
    Object.defineProperty(IntervalToken.prototype, "timeToSync", {
        get: function () {
            // no sync date; return true so it is synced
            if (!this.lastSyncedDate) {
                return true;
            }
            return moment(new Date()).subtract(this.syncInterval, "minutes").isAfter(this.lastSyncedDate);
        },
        enumerable: true,
        configurable: true
    });
    return IntervalToken;
}());
var LocalSync = /** @class */ (function () {
    function LocalSync(application, cache) {
        this.application = application;
        this.cache = cache;
        this.disposed = false;
        this.syncPromise = jQuery.Deferred().resolve().promise();
        // Make sure only one sync is in progress because we might get issues otherwise
        // (Data getting added twice because of race conditions)
        this.syncInProgress = false;
        this.syncBookKeeping = {};
        this.applicationDB = application.getDatabase();
        this.applicationDB.onerror = function (err) { return console.error(err); };
    }
    LocalSync.prototype.editSyncInProgress = function (value) {
        this.syncInProgress = value;
        jQuery(document).trigger("tsfSync", value);
    };
    LocalSync.prototype.IsSyncing = function () {
        return this.syncInProgress;
    };
    LocalSync.prototype.initialize = function () {
        var _this = this;
        var deferred = jQuery.Deferred();
        var syncStore = this.openSyncGroupLastSyncStore();
        var application = this.application;
        application.syncGroups.forEach(function (syncGroup) {
            _this.syncBookKeeping[syncGroup.syncGroupId] = new IntervalToken(syncGroup.syncInterval);
        });
        syncStore.then(function (store) {
            store.openCursor().onsuccess = function (event) {
                var cursor = event.target.result;
                if (cursor) {
                    var value = cursor.value;
                    // Guard against old groups which are in the local db but don't exist anymore
                    if (_this.syncBookKeeping[value.sync_group_id]) {
                        _this.syncBookKeeping[value.sync_group_id].lastSyncedDate = value.lastSuccessfulSync;
                    }
                    cursor.continue();
                    deferred.resolve();
                }
            };
            store.transaction.oncomplete = function () {
                deferred.resolve();
            };
        }, function () {
            deferred.reject();
        });
        return deferred.promise();
    };
    LocalSync.prototype.getSynchronisationPromise = function () {
        return this.syncPromise;
    };
    LocalSync.prototype.startLocalSyncInterval = function () {
        var _this = this;
        var interval = this.application.syncInterval * 60 * 1000; // Convert minutes to ms
        // Guard against starting the timer multiple times.
        clearTimeout(this.syncIntervalTimer);
        {
            var syncGroupsToSync = [];
            var syncGroups = this.application.syncGroups;
            for (var groupToken in this.syncBookKeeping) {
                var group = this.syncBookKeeping[groupToken];
                if (group.timeToSync && syncGroups.key(groupToken)) {
                    syncGroupsToSync.push(syncGroups.key(groupToken));
                }
            }
            if (syncGroupsToSync.length > 0) {
                this.fullSynchronize(false, syncGroupsToSync);
            }
        }
        var func = function () {
            try {
                var syncGroupsToSync_1 = [];
                _this.application.syncGroups.forEach(function (syncGroup) {
                    var intervalToken = _this.syncBookKeeping[syncGroup.syncGroupId];
                    if (intervalToken && intervalToken.timeToSync) {
                        syncGroupsToSync_1.push(syncGroup);
                    }
                });
                _this.application.serverSync.getQueue(true).then(function (items) {
                    if (items.length > 0 || syncGroupsToSync_1.length > 0) {
                        _this.fullSynchronize(false, syncGroupsToSync_1).always(function () {
                            // Make sure to not restart the timer when disposed, an user might have logged out while localsync was busy.
                            if (_this.disposed) {
                                return;
                            }
                            _this.syncIntervalTimer = window.setTimeout(func, interval);
                        });
                    }
                    else {
                        _this.syncIntervalTimer = window.setTimeout(func, interval);
                    }
                });
            }
            catch (err) {
                Utils.reportError(err);
                _this.syncIntervalTimer = window.setTimeout(func, interval);
            }
        };
        this.syncIntervalTimer = window.setTimeout(func, interval);
    };
    LocalSync.prototype.stopLocalSyncInterval = function () {
        clearTimeout(this.syncIntervalTimer);
    };
    LocalSync.prototype.refreshSyncGroup = function (objView) {
        // Don't do a localsync on online subjects or when the gui is offline
        if (!objView.obj.syncGroupId || !AppConfig.online || this.syncInProgress) {
            return jQuery.Deferred().reject();
        }
        var syncGroup = this.application.syncGroups.key(objView.obj.syncGroupId);
        if (syncGroup && !syncGroup.atomic) {
            // first sync the obj which we actually requested; this makes it more usable for endusers
            var index = syncGroup.objs.array.indexOf(objView.obj);
            if (index !== -1) {
                syncGroup.objs.array.splice(index, 1);
                syncGroup.objs.array.unshift(syncGroup.objs.object[objView.obj.objId]);
            }
        }
        if (syncGroup) {
            return this.fullSynchronize(false, [syncGroup]);
        }
        else {
            return jQuery.Deferred().reject();
        }
    };
    LocalSync.prototype.openLocalSyncTable = function (mode) {
        var deferred = jQuery.Deferred();
        if (!this.cache || this.cache.disposed) {
            return deferred.reject().promise();
        }
        if (!this.cache.database || this.disposed) {
            return deferred.reject().promise();
        }
        this.cache.database.onerror = function (err) {
            deferred.reject(err).promise();
        };
        var transaction = this.cache.database.transaction([TSFCache.localSyncTable], mode);
        return deferred.resolve(transaction.objectStore(TSFCache.localSyncTable)).promise();
    };
    LocalSync.prototype.loadSyncToken = function (obj) {
        var _this = this;
        var deferred = jQuery.Deferred();
        this.openLocalSyncTable("readonly").then(function (store) {
            var index = store.index(TSFCache.localSyncObjName);
            var request = index.openCursor(IDBKeyRange.only(_this.application.guiApplId + obj.objId));
            request.onsuccess = function () {
                var cursor = request.result;
                if (cursor) {
                    deferred.resolve(cursor.value["SyncToken"]);
                }
                else {
                    // No synctoken found
                    deferred.resolve();
                }
            };
            request.onerror = function () {
                deferred.reject();
            };
        }, function () { deferred.reject(); });
        return deferred.promise();
    };
    LocalSync.prototype.removeSyncToken = function (obj) {
        var _this = this;
        var deferred = jQuery.Deferred();
        var store = this.openLocalSyncTable("readwrite").then(function (store) {
            var index = store.index(TSFCache.localSyncObjName);
            var req = index.openCursor(IDBKeyRange.only(_this.application.guiApplId + obj.objId));
            req.onsuccess = function () {
                var cursor = req.result;
                if (cursor) {
                    cursor['delete']();
                }
                deferred.resolve();
            };
            req.onerror = function () {
                deferred.reject();
            };
        }, function () {
            deferred.reject();
        });
        return deferred.promise();
    };
    LocalSync.prototype.storeSyncToken = function (obj, syncToken) {
        var _this = this;
        var deferred = jQuery.Deferred();
        var store = this.openLocalSyncTable("readwrite").then(function (store) {
            var index = store.index(TSFCache.localSyncObjName);
            var key = _this.application.guiApplId + obj.objId;
            var request = index.openCursor(IDBKeyRange.only(key));
            request.onsuccess = function () {
                var cursor = request.result;
                if (cursor) {
                    cursor.value["SyncToken"] = syncToken;
                    cursor.value[TSFCache.localSyncObjName] = key;
                    cursor.update(cursor.value);
                }
                else {
                    var valueObject = {
                        ApplId: _this.application.guiApplId,
                        ObjId: obj.objId,
                        ApplIdObj: key,
                        SyncToken: syncToken
                    };
                    // Not found, add it now
                    store.add(valueObject);
                }
                deferred.resolve();
            };
            request.onerror = function () {
                deferred.reject();
            };
        }, function () { deferred.reject(); });
        return deferred.promise();
    };
    LocalSync.resetSyncTokens = function (applicationId) {
        var deferred = jQuery.Deferred();
        var cache = Model.TSFApplicationManager.Current.TSFCache;
        if (!cache)
            return deferred.reject().promise();
        var transaction = cache.database.transaction([TSFCache.localSyncTable], "readwrite"), store = transaction.objectStore(TSFCache.localSyncTable), index = store.index(TSFCache.localSyncIndexName);
        var req = index.openCursor(IDBKeyRange.only(applicationId));
        req.onsuccess = function () {
            var cursor = req.result;
            if (cursor) {
                cursor['delete']();
                cursor['continue']();
            }
        };
        store.transaction.oncomplete = function () {
            deferred.resolve();
        };
        return deferred.promise();
    };
    LocalSync.prototype.fullSynchronize = function (showProgress, syncGroupsToSync) {
        var _this = this;
        // hack for vdi; because there offline code has a bug where it calls this method with false, false causing it to crash
        // They can fix it but we don't want them to loose there offline changes so we decided to temporary fix it here
        if (typeof syncGroupsToSync == "boolean") {
            syncGroupsToSync = null;
        }
        var deferred = jQuery.Deferred();
        // Don't try to sync when for example the gui is offline but this code is called from the offline code - which doesn't check this
        if (!AppConfig.online || !this.application.model.containsOfflineObjs) {
            return deferred.resolve().promise();
        }
        // Wait for any editmode to finish. This is important because a synchronization might fail.
        // The synchronization will then automatically redirect to the synchronize page, leaving a controller in editmode.
        this.application.waitForEditmode().always(function () {
            // Guard against multiple syncs
            if (_this.syncInProgress) {
                // Resolve the promise in case we hit this twice or the syncPromise State will stay pending
                deferred.resolve();
                return;
            }
            _this.editSyncInProgress(true);
            var appl = _this.application;
            // First synchronize all local changes to the server.
            var serverSyncPromise = _this.application.serverSync.syncQueueToServer()
                .then(function (error, changedGroups) {
                // Perform a localSync when items are synced or when its time.
                if (error == 1 /* Network */) {
                    _this.editSyncInProgress(false);
                    deferred.reject();
                    return;
                }
                // when empty everything is syced but when this contains items we need to add the changedGroups to it
                // because they did contain local changes
                // Make sure to not add them twice
                if (syncGroupsToSync && changedGroups) {
                    var _loop_1 = function (groupId) {
                        if (!syncGroupsToSync.some(function (group) { return group.syncGroupId == groupId; })) {
                            syncGroupsToSync.push(changedGroups[groupId]);
                        }
                    };
                    for (var groupId in changedGroups) {
                        _loop_1(groupId);
                    }
                }
                if (syncGroupsToSync == null) {
                    syncGroupsToSync = appl.syncGroups.array;
                }
                var objCount = syncGroupsToSync.reduce(function (prev, cur) { return prev += cur.objs.array.length; }, 0) * 2;
                var syncDataPromise = _this.synchronizeDataToLocalStorage(syncGroupsToSync);
                if (showProgress) {
                    syncDataPromise.progress(function (index) {
                        var percentage = Math.round(((index + 1) / objCount) * 100);
                        Utils.showLoader(appl.language.transl(Model.TranslKeys.downloading) + ': ' + percentage + '%');
                    }).always(function () {
                        Utils.hideLoader();
                    });
                }
                syncDataPromise.always(function () {
                    _this.editSyncInProgress(false);
                    deferred.resolve();
                });
            })
                .fail(function (err) {
                _this.editSyncInProgress(false);
                // Make sure to hide the loader - the one from the progress
                // so the user can actually use the popup
                Utils.hideLoader();
                // Only show the fail popup when there isn't an open popup.
                // We have the popup over popup problem on multiple occasions.
                // TODO: find a reasonable solution for this problem.
                if (!jQuery.mobile.activePage.is("#login")) {
                    if (jQuery(".ui-popup").length == 0) {
                        var lang = _this.application.language;
                        var popup = ConfirmationPopup.CreateOkPopup();
                        var promise = popup.initialize(lang.transl(Model.TranslKeys.synchronize), lang.transl(Model.TranslKeys.errorDuringSynchronization));
                        promise.always(function () {
                            _this.navigateToSyncPage(deferred);
                        });
                    }
                    else {
                        _this.navigateToSyncPage(deferred);
                    }
                }
                else {
                    Raven.captureMessage(Model.TranslKeys.errorDuringSynchronization + ": " + err);
                }
            });
            if (showProgress) {
                serverSyncPromise.progress(function (progress) {
                    var roundedProgress = Math.round(progress);
                    Utils.showLoader(_this.application.language.transl(Model.TranslKeys.synchronize) + ': ' + roundedProgress + "%");
                });
                deferred.always(function () {
                    Utils.hideLoader();
                });
            }
        });
        this.syncPromise = deferred.promise();
        return this.syncPromise;
    };
    LocalSync.prototype.navigateToSyncPage = function (deferred) {
        // Only navigate to the synchronize page when it isn't the current page.
        var parsedPath = jQuery.mobile.path.parseUrl(jQuery.mobile.activePage.data('url'));
        if (parsedPath.filename != "synchronize.html") {
            jQuery.mobile.pageContainer.pagecontainer("change", "synchronize.html");
        }
        deferred.reject();
    };
    /*
     * Restore unsynced modifications; they might have been removed (or overwritten) due the synchronization
     * The local queue values are more important because they will eventually get synced (might fail due to triggers though)
     */
    LocalSync.prototype.reSyncOfflineTables = function () {
        var _this = this;
        var deferred = jQuery.Deferred();
        this.application.serverSync.getQueue(true).done(function (queue) {
            // We only care about unsynced items and insert, delete and update types
            queue = queue.filter(function (item) {
                return item.actionType >= 0 /* Insert */ && item.actionType <= 2 /* Delete */;
            });
            var queueLength = queue.length;
            var processNextRow = function (index) {
                var next = function () {
                    if (index < queueLength - 1) {
                        processNextRow(++index);
                    }
                    else {
                        deferred.resolve();
                    }
                };
                var item = queue[index];
                var application = _this.application;
                var wrapper = application.getStorageContainer();
                var obj = application.model.getObj(item.objId);
                if (!wrapper || !obj) {
                    deferred.resolve();
                    return;
                }
                // We need to resolve identity's because the sync succeeded. That might mean once you do an insert and directly an update (for example; could have been a delete)
                // the inserted row is synced (and an identity returned); localsync will get the insert and add it to the queue.
                // the update the user did might have happened on the old row (unresolved identity) or the new row (resolved identity)
                // So it is important to use the right value
                application.serverSync.resolveIdentities(item, obj.defaultObjView)
                    .always(function () {
                    if (item.actionType == 2 /* Delete */) {
                        // Locally deleted unsynced items might have been readded by localsync, delete them again so they don't show up again
                        // Not being there is the right thing because you can't change them anymore (because the delete is earlier in the queue
                        wrapper.deleteRow(obj, TSFPkBookmark.fromObject(obj, item.pk), false).always(function () {
                            next();
                        });
                    }
                    else if (item.actionType == 0 /* Insert */) {
                        // When localsync didn't found the synctoken, the inserts are gone - we need to check if the rows are there and if they are not
                        // we just add them again
                        // We need to bypass the indexeddbwrapper though, because that will create a new identity plus
                        var db = wrapper.getDatabase();
                        var transaction = db.transaction([obj.objId], "readwrite");
                        var store_1 = transaction.objectStore(obj.objId);
                        var index_1 = store_1.index(IndexedDb.IndexedDbWrapper.pkIndexName);
                        var req = index_1.openCursor(IndexedDb.createPKKeyRangeFromBookmark(obj, TSFPkBookmark.fromObject(obj, item.pk)));
                        req.onsuccess = function (evt) {
                            var cursor = evt.target.result;
                            if (!cursor) {
                                store_1.add(item.pk);
                            }
                        };
                        transaction.onerror = transaction.oncomplete = function () {
                            next();
                        };
                    }
                    else if (item.actionType == 1 /* Update */) {
                        // If online did an update, we need to redo the update because the one in our queue is more important (it will get synced later)
                        wrapper.updateRow(obj, TSFPkBookmark.fromObject(obj, item.pk), TSFRowBookmark.fromObject(obj, item.pk), false).always(function () {
                            next();
                        });
                    }
                    else {
                        // Just to be safe, shouldn't happen though because of the filter above
                        next();
                    }
                });
            };
            if (queueLength > 0) {
                processNextRow(0);
            }
            else {
                deferred.resolve();
            }
        });
        return deferred.promise();
    };
    LocalSync.prototype.synchronizeDataToLocalStorage = function (syncGroupsToSync) {
        var _this = this;
        var application = this.application;
        var index = 0;
        var synchronized = jQuery.Deferred();
        var progressNotifier = {
            increment: function () { synchronized.notify(index++); }
        };
        var synchronisationPromise = syncGroupsToSync.reduce(function (prevGroupResult, syncGroup) {
            return Utils.chainPromise(prevGroupResult, function () {
                return _this.syncSyncGroupInternal(syncGroup, progressNotifier);
            });
        }, jQuery.Deferred().resolve());
        // Don't use the promise returned by Utils.chainPromise without wrapping because that will corrupt the state for the next one
        // we only want to update and mark the controllers dirty when it didn't fail
        // And deferred.always doesn't wait when you return a promise - only the `then` method does that
        // We don't use then here because we need the same code twice (onfullfill and onreject)
        Utils.chainPromise(synchronisationPromise, function () {
            if (_this.disposed) {
                synchronized.resolve();
                return null;
            }
            _this.reSyncOfflineTables().always(function () {
                synchronized.resolve();
            });
            return null;
        });
        return synchronized.promise();
    };
    LocalSync.prototype.syncSyncGroupInternal = function (syncGroup, progressNotifier) {
        var _this = this;
        // Return a rejected promise when LocalSync is disposed.
        // This will break the chain of promises, since a then call only continues on resolved promises.
        if (this.disposed) {
            return jQuery.Deferred().reject().promise();
        }
        var changedObjs = [];
        // declare here so we can aggregate errors for non-atomic sync groups
        var item;
        var syncPromise = syncGroup.objs.array.reduce(function (prevObjResult, obj) {
            return prevObjResult.then(function () {
                progressNotifier.increment();
                var objSync = jQuery.Deferred();
                _this.fetchObjChanges(_this.application, obj).then(function () {
                    // fetchObjChanges was sucessful, for atomic groups continue to the next one;
                    // for non-atomic we process the changes directly
                    if (syncGroup.atomic) {
                        objSync.resolve();
                    }
                    else {
                        // non-atomic group; process directly so it doesn't require all the other obj's in the group to finish
                        return _this.processObjChanges(obj).then(function (changed) {
                            if (changed) {
                                _this.markControllersDirty([obj.objId]);
                            }
                            progressNotifier.increment();
                            objSync.resolve();
                        }, function () {
                            objSync.reject();
                        });
                    }
                }, function (errors) {
                    var syncGroupLastSyncStore = _this.openSyncGroupLastSyncStore();
                    // When item already exists it is always a LastSyncErrorItem (the success case is handled below)
                    // and this is the error handler for fetchObjChanges
                    // when there are multiple items in a non-atomic group we might end up getting here multiple times
                    // but we still continue to the next item
                    if (errors) {
                        if (item) {
                            item.error += "\r\n" + obj.defaultObjView.objTransl.transl + ": " + errors.join("\r\n");
                        }
                        else {
                            item = {
                                sync_group_id: syncGroup.syncGroupId,
                                error: obj.defaultObjView.objTransl.transl + ": " + errors.join("\r\n")
                            };
                        }
                    }
                    if (syncGroup.atomic) {
                        syncGroupLastSyncStore.then(function (store) {
                            store.put(item);
                        }, function () {
                            objSync.reject();
                        });
                        objSync.reject();
                    }
                    else {
                        // Continue when an error occurs on a non-atomic group; so always use resolve
                        objSync.resolve();
                    }
                });
                return objSync.promise();
            });
        }, jQuery.Deferred().resolve());
        // When the group is atomic, we first fetch all the obj's localsync responses and will process them here
        // for non-atomic we process them directly after fetching
        if (syncGroup.atomic) {
            syncPromise = syncPromise.then(function () {
                return syncGroup.objs.array.reduce(function (prevObjResult, obj) {
                    return prevObjResult.then(function () {
                        progressNotifier.increment();
                        return _this.processObjChanges(obj).then(function (changed) {
                            if (changed) {
                                changedObjs.push(obj.objId);
                            }
                        });
                    });
                }, jQuery.Deferred().resolve());
            });
        }
        return syncPromise.then(function () {
            var syncGroupLastSyncStore = _this.openSyncGroupLastSyncStore();
            var syncedDate = new Date();
            // item might already exists when error occurs for tables in a non-atomic group
            // We still want to update the sync date in that case or it will keep syncing the group
            if (item) {
                item.lastSuccessfulSync = syncedDate;
                if (item.error) {
                    item.errorDate = syncedDate;
                }
            }
            else {
                item = { sync_group_id: syncGroup.syncGroupId, lastSuccessfulSync: syncedDate };
            }
            syncGroupLastSyncStore.then(function (store) {
                store.put(item);
            });
            // Update last synced date only when sucessful
            _this.syncBookKeeping[syncGroup.syncGroupId].lastSyncedDate = syncedDate;
        })
            .always(function () {
            _this.markControllersDirty(changedObjs);
            _this.application.getStorageContainer().clearResponseBufferStore();
        });
    };
    //Because the ErrorReason enum is an inline const, we have to translate the member names manually.
    LocalSync.prototype.translateErrorReason = function (reason) {
        switch (reason) {
            case 0 /* Unknown */:
                return "Unknown";
            case 1 /* JavaScript */:
                return "JavaScript";
            case 2 /* Server */:
                return "Server";
            case 3 /* Network */:
                return "Network";
        }
        return "Unknown";
    };
    LocalSync.prototype.fetchObjChanges = function (application, obj) {
        var _this = this;
        var deferred = jQuery.Deferred();
        this.loadSyncToken(obj).always(function (syncToken) {
            var builder = RequestBuilderFactory.createOnlineRequestBuilder(application, obj.defaultObjView);
            builder.setCallbackMethod(function (reader, status) {
                var objId = obj.objId;
                if (!status.success || reader.length == 0) {
                    Logger.log("Failed to sync: " + objId +
                        " because of a " + _this.translateErrorReason(status.reason) + " error." +
                        "[" + status.messages.join(", ") + "]", LogType.ERROR);
                    deferred.reject(status.messages);
                    return;
                }
                // There should be only one item.
                var item = reader.item(0);
                var newSyncToken = item.getString("SyncToken");
                var delta = item.getBool("Delta");
                // If the new synctoken is the same as the old one, there is no point in queueing it to local storage.
                if (delta && newSyncToken == syncToken) {
                    Logger.log("Same token " + objId, LogType.SYNC);
                    // No changes is a valid result.
                    deferred.resolve();
                }
                else {
                    // Serialize and persist the response.
                    var xmlObjChanges_1 = reader.serialize();
                    var responseBufferStore = _this.openResponseBufferStore();
                    responseBufferStore.then(function (store) {
                        store.add({ obj_id: objId, xmlResponseData: xmlObjChanges_1 });
                        deferred.resolve();
                    }, function () {
                        deferred.reject();
                    });
                }
            });
            // Execute localSync.
            builder.localSync(syncToken);
        });
        return deferred.promise();
    };
    LocalSync.prototype.processObjChanges = function (obj) {
        var _this = this;
        var deferred = jQuery.Deferred();
        if (this.disposed) {
            return deferred.reject().promise();
        }
        this.loadSyncToken(obj).always(function (syncToken) {
            _this.getXmlObjResponseData(obj.objId).then(function (xmlObjResponseData) {
                var item = IndexedDb.IndexedDbArrayReader.deserialize(xmlObjResponseData).item(0);
                var newSyncToken = item.getString("SyncToken"), delta = item.getBool("Delta"), addedContracts = item.list("AddedContracts");
                // If the new synctoken is the same as the old one, there is nothing to do.
                // This shouldn't occur however, as these results will not be queued to the localstorage.
                if (delta && newSyncToken == syncToken) {
                    Logger.log("Same token " + obj.objId, LogType.SYNC);
                    deferred.resolve(false);
                }
                else if (delta) {
                    // new token is different, check the AddedContracts, ModifiedContracts and DeletedContracts
                    var modifiedContracts_1 = item.list("ModifiedContracts"), deletedRefContracts_1 = item.list("DeletedReferenceContracts");
                    Logger.log("Delta: " + obj.objId + " Adding " + (addedContracts ? addedContracts.length : 0) +
                        " Modified: " + (modifiedContracts_1 ? modifiedContracts_1.length : 0) +
                        " Deleted: " + (deletedRefContracts_1 ? deletedRefContracts_1.length : 0), LogType.SYNC);
                    // Call addData with tryUpdate true, because the data might already be in the indexedDB
                    _this.addData(obj, addedContracts, true)
                        .then(function () { return _this.addData(obj, modifiedContracts_1, true); })
                        .then(function () { return _this.deleteContracts(obj, deletedRefContracts_1); })
                        .always(function () {
                        _this.storeSyncToken(obj, newSyncToken).then(function () {
                            deferred.resolve(true);
                        }, function () { deferred.reject(); });
                    });
                }
                else {
                    // Old synctoken wasn't found (or not given), we have a new synctoken.
                    // Also, we need to clear the whole table and store the new data (only AddedContracts)
                    var logMessage_1 = null;
                    if (syncToken) {
                        logMessage_1 = "Old synctoken not found (by scl): " + obj.objId + " ";
                    }
                    else {
                        logMessage_1 = "Synctoken not found (local): " + obj.objId + " ";
                    }
                    if (!_this.applicationDB || _this.disposed) {
                        return deferred.reject().promise();
                    }
                    var transaction = _this.applicationDB.transaction([obj.objId], "readwrite"), objectStore = transaction.objectStore(obj.objId);
                    objectStore.clear().onsuccess = function () {
                        Logger.log(logMessage_1 + "Adding " + (addedContracts ? addedContracts.length : 0), LogType.SYNC);
                        _this.addData(obj, addedContracts, false).done(function () {
                            _this.storeSyncToken(obj, newSyncToken).then(function () {
                                deferred.resolve(true);
                            }, function () { deferred.reject(); });
                        });
                    };
                }
            }, function (xmlObjResponseData) {
                deferred.resolve(false);
            });
        });
        return deferred.promise();
    };
    LocalSync.prototype.openResponseBufferStore = function () {
        var deferred = jQuery.Deferred();
        var responseBufferTableName = IndexedDb.IndexedDbWrapper.responseBufferTableName;
        if (!this.applicationDB || this.disposed)
            return deferred.reject().promise();
        var transaction = this.applicationDB.transaction(responseBufferTableName, "readwrite");
        return deferred.resolve(transaction.objectStore(responseBufferTableName)).promise();
    };
    LocalSync.prototype.getXmlObjResponseData = function (objId) {
        var deferred = jQuery.Deferred();
        var responseBufferStore = this.openResponseBufferStore();
        responseBufferStore.then(function (store) {
            var request = store.get(objId);
            request.onsuccess = function (event) {
                if (request.result && request.result.xmlResponseData) {
                    deferred.resolve(request.result.xmlResponseData);
                }
                else {
                    deferred.reject();
                }
            };
            // Just to ensure covering all code paths, there's no obvious reason for this to ever fire.
            request.onerror = function (event) {
                deferred.reject();
            };
        }, function () { deferred.reject(); });
        return deferred.promise();
    };
    LocalSync.prototype.openSyncGroupLastSyncStore = function () {
        var deferred = jQuery.Deferred();
        if (!this.applicationDB || this.disposed) {
            return deferred.reject().promise();
        }
        var syncGroupLastSyncTableName = IndexedDb.IndexedDbWrapper.syncGroupLastSyncTableName;
        var transaction = this.applicationDB.transaction(syncGroupLastSyncTableName, "readwrite");
        return deferred.resolve(transaction.objectStore(syncGroupLastSyncTableName)).promise();
    };
    LocalSync.prototype.getSyncGroupLastSuccessfulSync = function (syncGroupId) {
        var syncGroupToken = this.syncBookKeeping[syncGroupId];
        return (syncGroupToken && syncGroupToken.lastSyncedDate);
    };
    LocalSync.prototype.markControllersDirty = function (objIds) {
        if (objIds.length == 0) {
            return;
        }
        // Get the list of master controller for the current application.
        var list = TSFController.masterControllers[this.application.guiApplId];
        var markDirtyFunc = function (controller) {
            // Recursively check the controller tree for matching objIds.
            // Stop the recursion if a match is found because all child controllers
            // of a match will also automatically be marked dirty.
            if (objIds.indexOf(controller.obj.objId) > -1) {
                // Only mark the controller dirty when it actually uses offline data.
                if (controller.objView.offlineAvailable) {
                    controller.markDirtyBySync();
                }
            }
            else {
                for (var i = 0; i < controller.childControllers.length; i++) {
                    var childController = controller.childControllers[i];
                    if (objIds.indexOf(childController.obj.objId) > -1) {
                        // Only mark the controller dirty when it actually uses offline data.
                        if (childController.objView.offlineAvailable) {
                            childController.markDirtyBySync();
                        }
                    }
                    else {
                        markDirtyFunc(childController);
                    }
                }
            }
        };
        // Iterate over all master controller and mark their child controller dirty if the objId matches.
        for (var controllerId in list) {
            var masterController = list[controllerId];
            markDirtyFunc(masterController);
        }
    };
    LocalSync.prototype.getFileColumns = function (obj) {
        var colsToDownload = [];
        var objProps = obj.objProps.array;
        for (var p = 0; p < objProps.length; p++) {
            var prop = objProps[p];
            if (prop.isFileColumn) {
                colsToDownload.push(prop);
            }
        }
        return colsToDownload;
    };
    LocalSync.prototype.deleteContracts = function (obj, data) {
        var _this = this;
        if (!data || data.length == 0) {
            return jQuery.Deferred().resolve().promise();
        }
        var deferred = jQuery.Deferred();
        if (!this.applicationDB || this.disposed) {
            return deferred.reject().promise();
        }
        var transaction = this.applicationDB.transaction(obj.objId, "readwrite");
        var store = transaction.objectStore(obj.objId);
        var fileColumns = this.getFileColumns(obj);
        var removableFileReferences = [];
        var rowPromise = jQuery.Deferred().resolve().promise();
        for (var d = 0; d < data.length; d++) {
            var rowItem = data.item(d);
            rowPromise = rowPromise.then((function (row) {
                return function () {
                    var def = jQuery.Deferred();
                    var index = store.index(IndexedDb.IndexedDbWrapper.pkIndexName);
                    var range = IndexedDb.createPKKeyRangeFromBookmark(obj, TSFPkBookmark.fromReader(obj, row));
                    var req = index.openCursor(range);
                    req.onsuccess = function () {
                        var cursor = req.result;
                        if (cursor) {
                            var indexedDbRow = cursor.value;
                            cursor['delete']();
                            for (var i = 0; i < fileColumns.length; i++) {
                                var col = indexedDbRow[fileColumns[i].objPropId];
                                if (col && col.HashCode) {
                                    removableFileReferences.push(_this.application.fileCache.removeFileReference(col));
                                }
                            }
                            def.resolve();
                        }
                        else {
                            def.resolve();
                        }
                    };
                    return def.promise();
                };
            })(rowItem));
        }
        rowPromise.then(function () {
            if (removableFileReferences.length > 0) {
                jQuery.when.apply(undefined, removableFileReferences).always(function () {
                    deferred.resolve();
                });
            }
            else {
                deferred.resolve();
            }
        }, function () {
            deferred.resolve(); // change to reject if it doens't fail?
        });
        return deferred.promise();
    };
    LocalSync.prototype.addData = function (obj, data, tryUpdate) {
        var colsToDownload = null;
        var dataLength = data ? data.length : 0;
        if (dataLength == 0) {
            return jQuery.Deferred().resolve().promise();
        }
        if (this.application.extendedProperties.downloadFiles && (colsToDownload = this.getFileColumns(obj)).length > 0) {
            return this.addDataWithImages(obj, data, colsToDownload, tryUpdate);
        }
        else {
            return this.addDataWithoutImages(obj, data, tryUpdate);
        }
    };
    LocalSync.prototype.addDataWithImages = function (obj, data, fileColumns, tryUpdate) {
        var _this = this;
        var deferred = jQuery.Deferred(), dataList = [];
        this.executeSerial(data, function (item, rowIndex, next) {
            var resolveList = [], normalizedRow = Utils.normalizeRow(item, obj.objProps.array), that = _this;
            for (var p = 0; p < fileColumns.length; p++) {
                var prop = fileColumns[p];
                resolveList.push((function (fileProp) {
                    return function () {
                        var fileDeferred = jQuery.Deferred();
                        var fileCache = that.application.fileCache;
                        if (!that.disposed) {
                            if (tryUpdate) {
                                fileCache.updateFileCache(obj, fileProp, normalizedRow).always(function () {
                                    fileDeferred.resolve();
                                });
                            }
                            else {
                                if (!Utils.isEmpty(normalizedRow[fileProp.objPropId])) {
                                    // Insert the file in tsf_mobile_files
                                    fileCache.addFileReference(obj, fileProp, normalizedRow).always(function () {
                                        fileDeferred.resolve();
                                    });
                                }
                                else {
                                    // use setTimeout or we might blow up the callstack when there aren't any rows with images.
                                    window.setTimeout(function () {
                                        fileDeferred.resolve();
                                    }, 0);
                                }
                            }
                        }
                        else {
                            fileDeferred.reject();
                        }
                        return fileDeferred.promise();
                    };
                })(prop));
            }
            return resolveList.reduce(function (prev, curr) {
                return prev.then(curr);
            }, jQuery.Deferred().resolve()).then(function () {
                dataList.push(normalizedRow);
                next();
            });
        }, function () {
            if (!tryUpdate) {
                _this.bulkInsert(dataList, obj, deferred);
            }
            else {
                _this.bulkUpdate(dataList, obj, deferred);
            }
        });
        return deferred.promise();
    };
    LocalSync.prototype.executeSerial = function (reader, iterator, finish) {
        var cycle = function (i) {
            if (i < reader.length) {
                iterator(reader.item(i), i, function () {
                    cycle(i + 1);
                });
            }
            else {
                finish();
            }
        };
        cycle(0);
    };
    LocalSync.prototype.addDataWithoutImages = function (obj, data, tryUpdate) {
        var deferred = jQuery.Deferred();
        var dataList = [];
        var dataLength = data.length;
        for (var i = 0; i < dataLength; i++) {
            var normalizedRow = Utils.normalizeRow(data.item(i), obj.objProps.array);
            dataList.push(normalizedRow);
        }
        if (!tryUpdate) {
            this.bulkInsert(dataList, obj, deferred);
        }
        else {
            this.bulkUpdate(dataList, obj, deferred);
        }
        return deferred.promise();
    };
    LocalSync.prototype.bulkInsert = function (dataList, obj, deferred) {
        if (!this.applicationDB || this.disposed) {
            return deferred.reject();
        }
        var transaction = this.applicationDB.transaction(obj.objId, "readwrite"), store = transaction.objectStore(obj.objId);
        // Directly on application database for high performance
        // Don't use the storageContainer.insertBookmark
        for (var d = 0; d < dataList.length; d++) {
            var rowObject = dataList[d];
            IndexedDb.addIndexes(obj, rowObject);
            store.add(rowObject);
        }
        transaction.oncomplete = transaction.onerror = function () {
            deferred.resolve();
        };
    };
    LocalSync.prototype.bulkUpdate = function (dataList, obj, deferred) {
        if (!this.applicationDB || this.disposed) {
            return deferred.reject();
        }
        var transaction = this.applicationDB.transaction(obj.objId, "readwrite"), store = transaction.objectStore(obj.objId);
        var rowPromise = jQuery.Deferred().resolve().promise();
        for (var d = 0; d < dataList.length; d++) {
            rowPromise = rowPromise.then((function (row) {
                return function () {
                    var def = jQuery.Deferred();
                    IndexedDb.addIndexes(obj, row);
                    var index = store.index(IndexedDb.IndexedDbWrapper.pkIndexName);
                    var range = IndexedDb.createPKKeyRangeFromBookmark(obj, TSFPkBookmark.fromObject(obj, row));
                    var req = index.openCursor(range);
                    req.onsuccess = function () {
                        var cursor = req.result;
                        if (cursor) {
                            row["tsf_id"] = cursor.primaryKey;
                            cursor.update(row);
                        }
                        else {
                            // No cursor found, add the row and resolve
                            store.add(row);
                        }
                        def.resolve();
                    };
                    return def.promise();
                };
            })(dataList[d]));
        }
        transaction.oncomplete = transaction.onerror = function () {
            deferred.resolve();
        };
    };
    LocalSync.prototype.dispose = function () {
        // Do not dispose the cache database here, it isn't handled by LocalSync.
        this.disposed = true;
        this.stopLocalSyncInterval();
        return this.syncPromise;
    };
    return LocalSync;
}());
